from __future__ import annotations

from typing import TypeVar, Generic, overload, NoReturn, Literal

from cohdl._core import (
    Entity,
    Bit,
    BitVector,
    Unsigned,
    Signal,
    Port,
    expr_fn,
    AssignMode,
)

from ._assignable_type import AssignableType
from ._core_utility import Value, Ref, nop
from ._context import Duration, Context, SequentialContext

T = TypeVar("T")
U = TypeVar("U")
N = TypeVar(
    "N", Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 16, 20, 24, 32, 64, 100, 128]
)

def add_entity_port(entity: Entity, port: Port, name: str | None = None) -> Port:
    """
    Adds a new port to the given entity.
    """

class Serialized(Generic[T], AssignableType):
    _elemtype_: type[T]

    @classmethod
    def from_raw(cls, raw: BitVector) -> Serialized[T]:
        """
        Create a new instance of `Serialized[T]` that uses `raw`
        as the internal, serialized representation of an object
        of type `T`.
        """

    def _assign_(self, source, mode: AssignMode) -> None: ...
    def value(self, qualifier=Value) -> T:
        """
        Converts the internal, serialized representation into an object of
        type `T` using `std.from_bits` and the optional type qualifier.
        """

    @property
    def ref(self) -> T:
        """
        Converts the internal, serialized representation into an object of
        type `T`. The elements of the returned object are references to
        bits in the serialized representation.

        This only works for types, that are trivially serializable meaning
        each member directly maps to a subrange of the serialized bits.
        Examples for trivially serializable types are
        the builtins Bit/BitVector/Signed/Unsigned, std.SFixed, std.UFixed
        and all std.Record types.

        `self.ref` is equivalent to `self.value(qualifier=std.Ref)`.
        `ref` is implemented as a property to allow its use in assignments.
        """

    def bits(self) -> BitVector:
        """
        Returns the internal, serialized representation.
        My be converted to an instance of `T` using std.from_bits.
        Objects of type `T` can be serialized and assigned to these bits.

        >>> def example(a, b):
        >>>     my_serialized = Signal[Serialized[MyType]](a, b)
        >>>     my_deserialized_1 = my_serialized.value()
        >>>     my_deserialized_2 = std.from_bits[MyType](my_serialized.bits())
        """

def as_readable_vector(*parts: Bit | BitVector) -> Signal[BitVector]:
    """
    Concatenates all arguments to a single BitVector.
    The resulting vector is not writeable because it is assigned
    in the function.

    This function cannot be used in synthesizable contexts.

    >>> def architecture(self):
    >>>     a = Signal[Bit]()
    >>>     b = Signal[BitVector[5]]()
    >>>     c = Signal[Bit]()
    >>>
    >>>     vec = std.as_readable_vector(a, b, c)
    >>>
    >>>     @std.concurrent
    >>>     def logic():
    >>>         assert c == vec[0]
    >>>         assert b == vec[5:1]
    >>>         assert a == vec[6]
    """

def as_writeable_vector(
    *parts: Signal[Bit] | Signal[BitVector], default=None
) -> Signal[BitVector]:
    """
    Takes a list of Bits and BitVectors and creates a new
    BitVector with the total width of all parts.

    Each part is assigned the state of the corresponding
    subsection of this new vector.

    `default` defines the initial state of the new BitVector.

    This function cannot be used in synthesizable contexts.

    >>> def architecture(self):
    >>>     a = Signal[Bit]()
    >>>     b = Signal[BitVector[5]]()
    >>>     c = Signal[Bit]()
    >>>
    >>>     vec = std.as_writeable_vector(a, b, c)
    >>>
    >>>     @std.concurrent
    >>>     def logic():
    >>>         # update the states of a, b and c
    >>>         # by assigning to the corresponding ranges in vec
    >>>         vec[0] <<= Null
    >>>         vec[5:1] <<= Full
    >>>         vec[6] <<= True
    >>>
    >>>         assert c == Bit(Null)
    >>>         assert b == BitVector[5](Full)
    >>>         assert a == Bit(True)
    """

class Array(Generic[T, N]):
    """
    A wrapper utility around cohdl.Array.
    Unlike the builtin cohdl.Array the element type of std.Array is not
    limited to builtins. Any serializable type can be stored.
    """

    def __init__(self, val=None, _qualifier_=Signal): ...
    def __len__(self) -> int:
        """
        returns the number of elements in the Array
        """

    def __getitem__(self, index: Unsigned | int) -> T:
        """
        Returns the element at the given index.

        This only works for trivially serializable element types,
        because internally `std.from_bits[T](raw, std.Ref)` is used
        to obtain the returned value.

        Use `self.get_elem` for other types.
        """

    def get_elem(self, index: Unsigned | int, qualifier=Ref) -> T:
        """
        Returns the value obtained by deserializing the stored element
        at `index` using the given `qualifier`.
        """

    def set_elem(self, index: Unsigned | int, value: T):
        """
        Assign the provided `value` to the element at `index`.
        """

#
#
#

def stringify(*args, sep: str = ""):
    """
    returns `sep.join(str(arg) for arg in args)`

    This function exists to allow basic string conversions
    in evaluated contexts.
    """

class DelayLine(Generic[T]):
    @overload
    def __init__(self, inp: T, delay: int):
        """
        DelayLine wraps a list of Signals of type T.
        With `delay + 1` elements. When used in a synthesizable context
        the value of each element is assigned to the next one
        every time the constructor is evaluated.

        `initial` defines the default value of the internal memory elements.

        DelayLines can be defined outside of synthesizable contexts,
        when the `ctx` argument is provided.

        >>> line_a = DelayLine(inp, 3, initial=Null ctx=seq_ctx)
        >>>
        >>> @seq_ctx
        >>> def example():
        >>>     line_b = DelayLine(inp, 5)
        >>>
        >>>     out.a <<= line_a.last() # inp delayed by 3
        >>>     out.a <<= line_a[1]     # inp delayed by 1
        >>>
        >>>     out.b <<= line_b.last() # inp delayed by 5
        >>>     out.b <<= line_b[2]     # inp delayed by 2
        """

    @overload
    def __init__(self, inp: T, delay: int, ctx: SequentialContext): ...
    @overload
    def __init__(self, inp: T, delay: int, initial): ...
    @overload
    def __init__(self, inp: T, delay: int, initial, ctx: SequentialContext): ...
    def __getitem__(self, delay: int) -> T: ...
    def __len__(self) -> int: ...
    def __iter__(self): ...
    def last(self) -> T: ...

@overload
def delayed(inp, delay: int, inital) -> Signal:
    """
    Returns a copy of `inp` delayed by a number of clock cycles.

    When `initial` is specified it is passed as the first argument
    of the constructors of the intermediate signals.

    >>>
    >>> @std.sequential(clk)
    >>> def example():
    >>>     out_a <<= std.delayed(inp, 3)
    >>>
    >>>     if update_b:
    >>>         out_b <<= std.delayed(inp, 2)
    >>>
    >>> # output:
    >>>
    >>> inp      : A B C D E F G H I J K L M
    >>> update_b : 0 1 1 1 1 1 0 0 0 1 1 1 1
    >>> ------------------------------------
    >>> out_a    : 0 0 0 A B C D E F G H I J
    >>> out_b    : 0 0 0 B C D D D D E F J K
    >>>
    """

@overload
def delayed(inp, delay: int) -> Signal: ...
def debounce(
    ctx: SequentialContext,
    inp: Signal[Bit],
    period: int | Duration,
    initial: bool | Bit = False,
    allowed_delta: float = 1e-9,
) -> Signal[Bit]:
    """
    Debounce the input signal using a saturating counter
    with a max-value defined by `period`.

    The counter is incremented, when `inp` is '1', and decremented
    when `inp` is '0'.

    The output signal is set to '1', when the max-value
    is reached, and to '0' when 0 is reached.

    On reset, the internal counter is initialized with `period/2`
    and the output is defined by `initial`.

    When `period` is defined as a Duration, `allowed_delta` defines the maximal
    allowed relative error in the debounce period.

    ---
    >>> # Example
    >>> def architecture(self):
    >>>     ctx = std.SequentialContext(self.clk)
    >>>     debounced = std.debounce(ctx, self.inp, 128)
    >>>
    >>>     # output debounced version of self.inp
    >>>     std.concurrent_assign(self.output, debounced)
    """

#
#
#

def max_int(arg: int | Unsigned) -> int:
    """
    returns the largest possible runtime value of `arg`
    - for Unsigned values this is 2**width-1
    - since integers are runtime constant they are returned unchanged
    """

def int_log_2(inp: int) -> int:
    """
    Returns the base two logarithm of a number.
    Asserts, that `inp` is of type int and a power of 2.
    """

def ceil_log_2(inp: int) -> int:
    """
    Return the logarithm to the base 2 of `inp` rounded to the next larger integer.
    """

def is_pow_two(inp: int):
    """
    check if `inp` is an integer power of two
    """

async def tick() -> None:
    """
    wait for a single clock cycle

    This will introduce an unconditional state transition.

    >>> # the following lines are equivalent
    >>> await std.tick()
    >>> await std.wait_for(1)
    >>> await cohdl.true
    """

@overload
async def wait_for(duration: int | Unsigned, *, allow_zero: bool = False) -> None:
    """
    wait for a given number of ticks

    To avoid a runtime check and logic duplication for the rarely needed
    case of waiting zero ticks, `duration` does not allow
    a value of 0. This can be changed by setting `allow_zero` to true.
    """

@overload
async def wait_for(duration: Duration) -> None:
    """
    wait for a given time duration

    wait_for uses `std.SequentialContext.current()` to determine the clock period
    of the enclosing synthesizable context and calculates the needed number
    of wait cycles from it. Because of that, this function can only be
    used in sequential contexts defined with a fixed frequency Clock.
    """

async def wait_forever() -> NoReturn:
    """
    stop execution of the coroutine

    This will introduce a statemachine state with no exit path.
    The only way to leave this function is a reset of the enclosing context.
    """

class Waiter:
    """
    Alternative to std.wait_for, that makes it possible to
    reuse the same counter register for multiple wait operations.
    """

    def __init__(self, max_duration: int | Duration):
        """
        Initialize a Waiter that can wait up to a given duration.
        """

    async def wait_for(self, duration: int | Duration):
        """
        same as std.wait_for() but all calls to this method, on
        a single instance of Waiter, use the same counter register.

        `duration` may not exceed the `max_duration` defined in the constructor.
        """

class OutShiftRegister:
    def __init__(self, src: BitVector, msb_first=False, unchecked=False):
        """
        Initializes a shift register with the current state of `src`.
        `msb_first` defines the order in which the bits will be
        shifted out of the register.

        When `unchecked` is set to True no assertions for shifts
        from empty registers are generated. Instead zeros will be
        shifted out. The methods `shift_all` and `empty` cannot be
        used when this option is set.
        """

    def set_data(self, data: BitVector):
        """
        Reinitialize the shift register with the given data
        `data` must have the same width as `src` in __init__

        Warning: setting data is a signal assignment
        and will be overwritten if `shift` is called
        after `set_data` in the same clock cycle.
        """

    async def shift_all(
        self, target: Signal[Bit] | Signal[BitVector], shift_delayed=False
    ):
        """
        On each clock cycle target is updated by shifting
        the corresponding number of bits out of the shift register.
        This coroutine returns once the register is empty.

        When `shift_delayed` is set to True shifting starts with
        a delay of one clock cycle.
        """

    def empty(self):
        """
        Returns True when all bits have been shifted out of the register
        """

    @overload
    def shift(self) -> Bit:
        """
        Shifts the register by one bit
        and returns the state of the shifted out bit.

        This method can only be called once per clock cycle
        and may not be mixed with `shift_all`.
        """

    @overload
    def shift(self, count: int) -> BitVector:
        """
        Shifts the register by `count` bits
        and returns a vector of the shifted out bits.

        This method can only be called once per clock cycle
        and may not be mixed with `shift_all`.
        """

class InShiftRegister:
    def __init__(self, len: int, msb_first=False, unchecked=False):
        """
        Initializes an empty shift register of length `len`.
        `msb_first` defines the order in which the bits will be
        shifted out of the register.

        When `unchecked` is set to True no assertions for shifts
        into full registers are generated. Instead old bits will be
        shifted out. The methods `shift_all` and `full` cannot be
        used when this options is set.
        """

    async def shift_all(self, src: Bit | BitVector, shift_delayed=False):
        """
        Shift the state of `src` into the shift register
        on each clock cycle until it is full.
        When `shift_delayed` is set to True shifting starts with
        a delay of one clock cycle.
        """

    def clear(self):
        """
        Reinitialize the shift register.

        Warning: clearing is a signal assignment
        and will be overwritten if `shift` is called
        after `clear` in the same clock cycle.
        """

    def full(self):
        """
        Returns True when `len` bits of data have been
        shifted into the register.
        """

    def shift(self, src: Bit | BitVector):
        """
        Shifts the content of `src` into the register.

        This method can only be called once per clock cycle
        and may not be mixed with `shift_all`.
        """

    def data(self):
        """
        Returns the deserialized data.
        This call is only valid when the register is full.
        """

def continuous_counter(
    ctx: Context, limit: int | Unsigned, *, on_change=nop
) -> Signal[Unsigned]:
    """
    Returns a unsigned signal that is incremented on each tick of `ctx`.
    When `limit` is reached the counter continues from zero.

    `on_change` is an optional callback function. It is invoked every time
    the counter value is updated with the new counter value as the only argument.

    example:

    >>> cnt = std.continuous_counter(ctx, 3)
    >>> # produces the sequence `0-1-2-3-0-1-2-...`
    """

class ToggleSignal:
    def __init__(
        self,
        ctx: Context,
        first_duration: int | Unsigned | Duration,
        second_duration: int | Unsigned | Duration | None = None,
        *,
        default_state: bool = False,
        first_state: bool = False,
        require_enable: bool = False,
        on_rising=None,
        on_falling=None,
    ):
        """
        Defines a bit signal that toggles between `0` and `1` with a defined
        period and duty cycle.

        The duration parameters define how long the signal remains in each state.
        These values are relative to the clock tick of `ctx`. `int` and `Unsigned`
        parameters are interpreted as a number of clock ticks. When a `std.Duration`
        is given, the number of ticks is inferred from the clock of `ctx`
        (so this only works if the clock has a defined frequency).
        `first_duration` and `second_duration` can have different types.

        If only `first_duration` is given the value will be reused for
        `second_duration` creating a signal with 50% duty cycle.

        `default_state` defines the signal state during reset and when
        the ToggleSignal is disabled.

        `first_state` defines the state of the signal when starting after a reset.
        I.e. `first_state=False` produces `XX0101...` whereas `first_state=True` produces `XX1010` where
        `X` is the reset state defined by `default_state`.

        By default the toggle signal starts as soon as the reset condition of `ctx` becomes false.
        When `require_enable` is set to True it will wait until `enable` is called.

        `on_rising` and `on_falling` are optional callback functions. They are invoked
        when ToggleSignal updates the internal state (i.e. in the clock cycle before
        the state change becomes visible to other sequential contexts).
        """

    def get_reset_signal(self) -> Signal[Bit]:
        """
        Returns the signal used to reset the internal toggle mechanism.
        `enable` and `disable` are helper methods that set/reset this signal.
        """

    def enable(self) -> None:
        """
        Start the toggle signal after a previous call to `disable`.
        """

    def disable(self) -> None:
        """
        Stop the toggle signal and reset the internal counter to zero.
        """

    def state(self) -> Signal[Bit]:
        """
        Returns the bit signal that is toggled by this instance of ToggleSignal.
        """

    def rising(self) -> Signal[Bit]:
        """
        Returns a bit signal that is `1` for a single clock cycle after
        each transition of the toggled signal from `0` to `1`.
        """

    def falling(self) -> Signal[Bit]:
        """
        Returns a bit signal that is `1` for a single clock cycle after
        each transition of the toggled signal from `1` to `0`.
        """

class ClockDivider:
    def __init__(
        self,
        ctx: SequentialContext,
        duration: int | Unsigned | Duration,
        *,
        default_state: bool = False,
        tick_at_start: bool = False,
        require_enable: bool = False,
        on_rising=None,
        on_falling=None,
    ):
        """
        ClockDivider takes a SequentialContext and a duration.
        It generates a signal that is high for one clock cycle
        and low for the rest of a period with the given duration.
        """

    def get_reset_signal(self) -> Signal[Bit]:
        """
        Returns the signal used to reset the internal counter mechanism.
        `enable` and `disable` are helper methods that set/reset this signal.
        """

    def enable(self) -> None:
        """
        Start signal generation after a previous call to `disable`.
        """

    def disable(self) -> None:
        """
        Stop signal generation and reset the internal counter to zero.
        """

    def state(self) -> Signal[Bit]:
        """
        Returns the bit signal that is toggled by this instance of ClockDivider.
        """

    def rising(self) -> Signal[Bit]:
        """
        Returns a bit signal that is `1` for a single clock cycle after
        each transition of self.state() from `0` to `1`.
        """

    def falling(self) -> Signal[Bit]:
        """
        Returns a bit signal that is `1` for a single clock cycle after
        each transition of self.state() from `1` to `0`.
        """

class SyncFlag:
    """
    SyncFlag is used to send single bit notifications between two clocked
    contexts.

    ---
    Example:

    >>> # in sender context
    >>> flag.set()
    >>> # optionally wait until flag is cleared by receiver
    >>> await flag.is_clear()
    >>>
    >>> # in receiver context
    >>> # wait until flag is set and clear it
    >>> await flag.receive()

    ---

    SyncFlag can be used as the argument of an `async with` statement.
    The process will wait until the flag becomes set and automatically clears it once
    the with-scope is left.

    >>> flag = SyncFlag()
    >>> async with flag:
    >>>     ...

    is equivalent to

    >>> flag = SyncFlag()
    >>> await flag.is_set()
    >>> ...
    >>> flag.clear()
    """

    def __init__(
        self,
        *,
        delay: int | None = None,
        rx_delay: int | None = None,
        tx_delay: int | None = None,
    ):
        """
        The optional delay parameters define delay lines for the two internal bit signals.

        When `delay` is set its value is applied to both `rx_delay` and `tx_delay`.

        When `rx_delay` is set to a value greater than zero, the signal set by
        the setter context, is delayed by the specified amount of clock cycles,
        before it is detected in the receiver context. The `rx_delay` is implemented
        in the receiver context.

        `tx_delay` is the opposite of `rx_delay` and specifies a synchronization
        delays for the `clear` operation. This delay is implemented in the sender context.
        """

    def set(self) -> None:
        """
        Set the flag. This has no effect when it is already set.

        This method is used in the sender context.
        """

    def clear(self) -> None:
        """
        Clear the flag. This has no effect  when it is already clear.

        This method should be used in the receiver context
        after a received `set` was processed.
        """

    @expr_fn
    def is_set(self) -> bool:
        """
        Check if the flag is set.

        Note: This method is marked as `expr_fn` and thus awaitable when used as the
        argument of an await expression.
        """

    @expr_fn
    def is_clear(self) -> bool:
        """
        Check if the flag is clear. (opposite of is_set)

        Note: This method is marked as `expr_fn` and thus awaitable when used as the
        argument of an await expression.
        """

    async def receive(self) -> None:
        """
        Wait until the SyncFlag is set and immediately clear it.
        """

    async def __aenter__(self) -> None: ...
    async def __aexit__(self, val, type, traceback) -> None: ...

class Mailbox(Generic[T]):
    """
    Mailbox combines a signal of data with a `std.SyncFlag`.

    A sender process uses the `send` method to write data and
    mark it as valid.

    ---
    Example:

    >>> # in common scope
    >>> mailbox = Mailbox[BitVector[8]]()
    >>>
    >>> # in sender context
    >>> mailbox.send(data_to_send)
    >>> # optionally wait until flag is cleared by receiver
    >>> await mailbox.is_clear()
    >>>
    >>> # in receiver context
    >>> # wait until flag is set and clear it
    >>> received_data = await flag.receive()
    """

    def __init__(
        self, *, delay: int = None, tx_delay: int = None, rx_delay: int = None
    ):
        """
        Create a Mailbox that can transmit data of the given `type`.
        Internally a `std.SyncFlag` is used to synchronize data access between
        a sending and a receiving context.
        The delay parameters are forwarded to that `std.SyncFlag` and allow for
        basic clock domain crossing.

        `tx_delay` specifies, after how many clock ticks the receiver context
        sees requests set by the sender. (relative to receiver context clock).

        `rx_delay` specifies after how many clock ticks the sender context
        sees acknowledge states set by the receiver. (relative to the sender context clock).
        """

    def send(self, data: T):
        """
        Put data into Mailbox and mark it as valid.
        """

    async def receive(self) -> T:
        """
        Wait until data becomes valid and return it.
        This method clears the valid flag.
        """

    def data(self) -> T:
        """
        Return the contained data.
        The content is only valid when self.is_set() return true.
        """

    @expr_fn
    def is_set(self) -> bool:
        """
        Returns true if the Mailbox contains data.
        """

    @expr_fn
    def is_clear(self) -> bool:
        """
        Returns true when the Mailbox is empty.
        """

    def clear(self) -> None:
        """
        Unset the internal flag. In the next clock cycle `is_clear` will return true.
        `is_set` will return true once the sender writes new data using `send`.
        """

#
#
#

class Fifo(Generic[T, N]):
    """
    A first-in-first-out container that can hold up to N-1 elements
    of type T.

    """

    def __init__(self, name: str): ...
    def push(self, data: T):
        """
        push one element onto the Fifo

        may only be called once per clock cycle
        may not be called on a full Fifo
        """

    def pop(self) -> T:
        """
        remove one element from the Fifo
        returns the removed element

        may not be called on an empty Fifo
        """

    def front(self) -> T:
        """
        Returns the state of element at the front of the Fifo
        i.e. the next value returned by `pop` without removing it.

        The result is undefined while the Fifo is empty
        """

    def empty(self) -> Bit:
        """
        check if fifo is empty
        """

    def full(self) -> Bit:
        """
        check if fifo is full
        """



from __future__ import annotations

from cohdl import BitVector, Null, Full

from ._core_utility import Value

from ._assignable_type import AssignableType

import enum

import typing
from typing import Self

T = typing.TypeVar("T")

#
#
#

class FixedRoundStyle(enum.Enum):
    ROUND = enum.auto()
    TRUNCATE = enum.auto()

class FixedOverflowStyle(enum.Enum):
    SATURATE = enum.auto()
    WRAP = enum.auto()

#
#

class _FixedResize(typing.Generic[T]):
    def __call__(
        self,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ) -> T: ...

class _Resize(typing.Generic[T]):
    def __call__(
        self,
        left: int,
        right: int,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ) -> T: ...
    def __getitem__(self, arg: slice) -> _FixedResize[T]: ...

#
#

class SFixed(AssignableType):
    resize: _Resize[SFixed]

    def resize_fn(
        self,
        left: int,
        right: int,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ) -> SFixed: ...
    def __class_getitem__(cls, arg: slice) -> SFixed: ...
    @classmethod
    def _count_bits_(cls) -> int: ...
    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier=Value) -> Self: ...
    def _to_bits_(self) -> BitVector: ...
    @classmethod
    def min(cls) -> float: ...
    @classmethod
    def max(cls) -> float: ...
    @classmethod
    def right(cls) -> int: ...
    @classmethod
    def left(cls) -> int: ...
    def __init__(
        self, val: int | Null | Full | SFixed = None, *, _qualifier_=Value
    ): ...
    def __eq__(self, other: int | float | SFixed) -> bool: ...
    def __bool__(self) -> bool: ...
    def __abs__(self) -> SFixed: ...
    def __add__(self, other: SFixed) -> SFixed: ...
    def __sub__(self, other: SFixed) -> SFixed: ...
    def __mul__(self, other: SFixed) -> SFixed: ...

#
#

class UFixed(AssignableType):
    resize: _Resize[UFixed]

    def resize_fn(
        self,
        left: int,
        right: int,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ) -> UFixed: ...
    def __class_getitem__(cls, arg: slice) -> UFixed: ...
    @classmethod
    def _count_bits_(cls) -> int: ...
    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier=Value) -> Self: ...
    def _to_bits_(self) -> BitVector: ...
    @classmethod
    def min(cls) -> int: ...
    @classmethod
    def max(cls) -> float: ...
    @classmethod
    def right(cls) -> int: ...
    @classmethod
    def left(cls) -> int: ...
    def __init__(
        self, val: int | Null | Full | SFixed = None, *, _qualifier_=Value
    ): ...
    def __eq__(self, other: int | float | UFixed) -> bool: ...
    def __bool__(self) -> bool: ...
    def __abs__(self) -> Self: ...
    def __add__(self, other: UFixed) -> UFixed: ...
    def __sub__(self, other: UFixed) -> UFixed: ...
    def __mul__(self, other: UFixed) -> UFixed: ...



_pretty_traceback = True


def use_pretty_traceback(val: bool = True):
    """
    CoHDL inserts stack frames from traced python functions
    into the exception trace back. This makes it a lot easier to debug
    compilation errors because the call stack contains the actual code,
    that caused the issue and not the compiler functions used to
    translate it.

    This mode can be switched of using `cohdl.use_pretty_traceback(False)`.

    You probably only want to set this value to False if you want to debug an internal compiler issue.
    """

    global _pretty_traceback
    _pretty_traceback = val


def pretty_traceback_active():
    return _pretty_traceback



from cohdl import consteval

pyeval = consteval


class _MetaClassInfo:
    def __init__(self, original_init):
        self._original_init = original_init


class MetaMember:
    def __class_getitem__(cls, args):
        class _MetaMember(cls):
            pass

        setattr(_MetaMember, "@meta_args", args)
        return _MetaMember

    def _meta_config_(self, parent, /, *args, **kwargs):
        pass


@pyeval
def _wrapped_init_(self, *args, **kwargs):
    meta_class_info: _MetaClassInfo = getattr("@meta_class_info", type(self))
    meta_class_info._original_init(self, *args, **kwargs)

    for name, ann in type(self).__annotations__.items():
        if not issubclass(ann, MetaMember):
            continue

        if hasattr(self, name):
            assert isinstance(getattr(self, name), ann)
        else:
            assert issubclass(ann, MetaMember)
            setattr(self, name, ann())

        obj: MetaMember = getattr(self, name)
        obj._meta_config_(self, name, *getattr(ann, "@meta_args"))


class MetaObject:
    def __init_subclass__(cls) -> None:
        original_init = cls.__init__
        setattr("@meta_class_info", _MetaClassInfo(original_init))
        cls.__init__ = _wrapped_init_



from __future__ import annotations

from cohdl import (
    BitVector,
    Signed,
    Unsigned,
    TypeQualifier,
    Null,
    Full,
    pyeval,
    AssignMode,
    static_assert,
)

from ._core_utility import Value, instance_check, choose_first

from ._assignable_type import AssignableType
from ._template import Template


import enum

import typing

T = typing.TypeVar("T")


#
#
#


class FixedRoundStyle(enum.Enum):
    ROUND = enum.auto()
    TRUNCATE = enum.auto()


class FixedOverflowStyle(enum.Enum):
    SATURATE = enum.auto()
    WRAP = enum.auto()


#
#


class _FixedResize(typing.Generic[T]):
    @pyeval
    def __init__(self, obj, left, right):
        self._obj = obj
        self.left = left
        self.right = right

    def __call__(
        self,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ) -> T:
        return self._obj.resize_fn(
            self.left,
            self.right,
            round_style=round_style,
            overflow_style=overflow_style,
        )


class _Resize(typing.Generic[T]):
    @pyeval
    def __init__(self, obj=None):
        self._obj = obj

    @pyeval
    def __get__(self, obj, objtype=None):
        return _Resize(obj)

    def __call__(
        self,
        left: int,
        right: int,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ) -> T:
        return self._obj.resize_fn(
            left, right, round_style=round_style, overflow_style=overflow_style
        )

    @pyeval
    def __getitem__(self, arg: slice) -> _FixedResize[T]:
        assert isinstance(arg, slice)
        assert arg.step is None, "slice step may not be used"
        return _FixedResize(self._obj, arg.start, arg.stop)


#
#


class _FixedTemplateArg:
    width: int
    exp: int

    def __init__(self, arg: slice):
        assert arg.step is None, "slice step may not be used"
        assert isinstance(arg.start, int), "slice start must be an integer"
        assert isinstance(arg.stop, int), "slice stop must be an integer"
        assert arg.start >= arg.stop, "slice start must be larger tan slice stop"

        self.arg = arg
        self.width: int = arg.start - arg.stop + 1
        self.exp: int = arg.stop

    def __hash__(self):
        return hash((self.width, self.exp))

    def __eq__(self, other):
        return self.arg == other.arg

    def __repr__(self):
        return f"[{self.arg.start}:{self.arg.stop}]"


class SFixed(Template[_FixedTemplateArg], AssignableType):
    _width: _FixedTemplateArg.width
    _exp: _FixedTemplateArg.exp

    @classmethod
    @pyeval
    def _count_bits_(cls):
        return cls._width

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier=Value):
        return cls(raw=bits.signed, _qualifier_=qualifier)

    def _to_bits_(self):
        return Value(self._val.bitvector)

    @classmethod
    @pyeval
    def min(cls):
        return -((2 ** (cls._width - 1)) * 2**cls._exp)

    @classmethod
    @pyeval
    def max(cls):
        return (2 ** (cls._width - 1) - 1) * 2**cls._exp

    @classmethod
    @pyeval
    def _adjust_val(cls, val):
        return int(val / 2**cls._exp)

    @pyeval
    def __repr__(self):
        val = TypeQualifier.decay(self._val).to_int() * 2**self._exp
        return f"{type(self)}({val})"

    @classmethod
    @pyeval
    def right(cls):
        return cls._exp

    @classmethod
    @pyeval
    def left(cls):
        return cls._exp + cls._width - 1

    def __init__(self, val=None, *, raw=None, _qualifier_=Value):
        raw_type = Signed[self._width]

        if raw is not None:
            assert val is None, "val and raw cannot be set at the same time"
            assert instance_check(raw, raw_type), "raw type must match underlying type"
            self._val: Signed = _qualifier_(raw)
        else:
            if val is None or val is Full or val is Null:
                self._val = _qualifier_[raw_type](val)
            elif isinstance(val, (int, float)):
                static_assert(
                    self.min() <= val <= self.max(),
                    "value outside valid range of fixed point number",
                )
                self._val = _qualifier_[raw_type](self._adjust_val(val))
            elif instance_check(val, Signed):
                zeros = -self._exp
                static_assert(zeros >= 0)

                self._val = _qualifier_(raw_type, val.resize(self._width, zeros=zeros))
            elif instance_check(val, Unsigned):
                zeros = -self._exp
                static_assert(zeros >= 0)
                self._val = _qualifier_[raw_type](
                    val.resize(self._width - 1, zeros=zeros)
                )
            elif instance_check(val, SFixed):
                if self.left() == val.left() and self.right() == val.right():
                    self._val = _qualifier_[raw_type](val._val)
                else:
                    assert self.left() >= val.left()
                    assert self.right() <= val.right()

                    zeros = self.right() - val.right()

                    self._val = _qualifier_[raw_type](
                        val._val.resize(self._width, zeros=zeros)
                    )
            else:
                raise AssertionError("invalid arg")

    #
    #
    #

    def _assign_(self, source, mode: AssignMode) -> None:
        if isinstance(source, SFixed):
            assert (
                source._width == self._width and source._exp == self._exp
            ), "source type ({}) does not match target type ({})".format(
                type(source), type(self)
            )

            self._val._assign_(source._val, mode)
        else:
            self._assign_(type(self)(source), mode)

    #
    #

    def __eq__(self, other: int | float | SFixed):
        if isinstance(other, (int, float)):
            return type(self)(other) == self
        else:
            assert isinstance(other, SFixed)
            assert type(other) is type(self)

            return self._val == other._val

    #
    #

    def __bool__(self):
        return bool(self._val)

    def __abs__(self):
        return type(self)(raw=abs(self._val))

    def __add__(self, other: SFixed):
        assert isinstance(other, SFixed)

        target_right = min(self.right(), other.right())
        target_left = max(self.left(), other.left()) + 1
        target_width = target_left - target_right + 1

        lhs_zeros = self.right() - target_right
        rhs_zeros = other.right() - target_right

        return SFixed[target_left:target_right](
            raw=self._val.resize(target_width, zeros=lhs_zeros)
            + other._val.resize(target_width, zeros=rhs_zeros)
        )

    def __sub__(self, other: SFixed):
        assert isinstance(other, SFixed)

        target_right = min(self.right(), other.right())
        target_left = max(self.left(), other.left()) + 1
        target_width = target_left - target_right + 1

        lhs_zeros = self.right() - target_right
        rhs_zeros = other.right() - target_right

        return SFixed[target_left:target_right](
            raw=self._val.resize(target_width, zeros=lhs_zeros)
            - other._val.resize(target_width, zeros=rhs_zeros)
        )

    def __mul__(self, other: SFixed):
        assert isinstance(other, SFixed)

        return SFixed[self.left() + other.left() + 1 : self.right() + other.right()](
            raw=self._val * other._val
        )

    resize = _Resize()

    def resize_fn(
        self,
        left: int,
        right: int,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ):
        selfleft = self.left()
        selfright = self.right()

        if selfleft == left and selfright == right:
            return type(self)(raw=Value(self._val))

        Result = SFixed[left:right]

        if selfleft > left:
            overflow = selfleft - left

            if overflow_style is FixedOverflowStyle.WRAP:
                if selfright >= right:
                    zeros = selfright - right

                    if overflow >= self._val.width:
                        return Result(raw=Signed[Result._width](Null))
                    else:
                        return Result(
                            raw=self._val.lsb(rest=overflow).signed.resize(zeros=zeros)
                        )
                else:
                    cutoff = right - selfright

                    if round_style is FixedRoundStyle.TRUNCATE:
                        return Result(
                            raw=Value[Signed[Result._width]](
                                self._val.lsb(rest=overflow).msb(rest=cutoff).signed
                            )
                        )
                    else:
                        assert (
                            round_style is FixedRoundStyle.ROUND
                        ), "invalid round_style {}".format(round_style)

                        if cutoff == 1:
                            do_round = (
                                Signed[2](1)
                                if self._val[cutoff - 1] and self._val[cutoff]
                                else Signed[2](0)
                            )
                        else:
                            do_round = (
                                Signed[2](1)
                                if self._val[cutoff - 1]
                                and (self._val[cutoff] or self._val[cutoff - 2 : 0])
                                else Signed[2](0)
                            )

                        return Result(
                            raw=Value[Signed[Result._width]](
                                self._val.lsb(rest=overflow).msb(rest=cutoff).signed
                                + do_round
                            )
                        )
            else:
                assert (
                    overflow_style is FixedOverflowStyle.SATURATE
                ), "invalid overflow_style {}".format(overflow_style)
                sign_bit = self._val.msb()

                overflow_bitcnt = min(selfleft - left, self._width - 1)
                overflow_bits = self._val.lsb(rest=1).msb(overflow_bitcnt)

                does_overflow = not sign_bit and overflow_bits
                does_underflow = sign_bit and ~overflow_bits

                if selfright >= right:
                    zeros = selfright - right

                    if self._val.width <= overflow:
                        default_bits = Signed[Result._width](0)
                    else:
                        default_bits = self._val.lsb(rest=overflow).signed.resize(
                            Result._width, zeros=zeros
                        )

                    return Result(
                        raw=Value[Signed[Result._width]](
                            choose_first(
                                (does_underflow, Signed[Result._width].min()),
                                (does_overflow, Signed[Result._width].max()),
                                default=default_bits,
                            )
                        )
                    )
                else:
                    cutoff = right - selfright

                    if round_style is FixedRoundStyle.TRUNCATE:
                        return Result(
                            raw=Value[Signed[Result._width]](
                                choose_first(
                                    (does_underflow, Signed[Result._width].min()),
                                    (does_overflow, Signed[Result._width].max()),
                                    default=self._val.lsb(rest=overflow)
                                    .msb(rest=cutoff)
                                    .signed,
                                )
                            )
                        )
                    else:
                        assert (
                            round_style is FixedRoundStyle.ROUND
                        ), "invalid round_style {}".format(round_style)

                        if cutoff == 1:
                            do_round = (
                                Signed[2](1)
                                if self._val[cutoff - 1] and self._val[cutoff]
                                else Signed[2](0)
                            )
                        else:
                            do_round = (
                                Signed[2](1)
                                if self._val[cutoff - 1]
                                and (self._val[cutoff] or self._val[cutoff - 2 : 0])
                                else Signed[2](0)
                            )

                        selected_bits = self._val.lsb(rest=overflow + 1).msb(
                            rest=cutoff
                        )
                        overflow_or_full = does_overflow or not ~selected_bits

                        return Result(
                            raw=Value[Signed[Result._width]](
                                choose_first(
                                    (does_underflow, Signed[Result._width].min()),
                                    (overflow_or_full, Signed[Result._width].max()),
                                    default=self._val.lsb(rest=overflow)
                                    .msb(rest=cutoff)
                                    .signed
                                    + do_round,
                                )
                            )
                        )

        else:
            if selfright >= right:
                return Result(
                    raw=self._val.resize(Result._width, zeros=selfright - right)
                )
            else:
                cutoff = right - selfright

                if round_style is FixedRoundStyle.TRUNCATE:
                    return Result(
                        raw=Value[Signed[Result._width]](
                            self._val.msb(rest=cutoff).signed.resize(Result._width)
                        )
                    )
                else:
                    assert (
                        round_style is FixedRoundStyle.ROUND
                    ), "invalid round_style {}".format(round_style)

                    if cutoff == 1:
                        do_round = (
                            Signed[2](1)
                            if self._val[cutoff - 1] and self._val[cutoff]
                            else Signed[2](0)
                        )
                    else:
                        do_round = (
                            Signed[2](1)
                            if self._val[cutoff - 1]
                            and (self._val[cutoff] or self._val[cutoff - 2 : 0])
                            else Signed[2](0)
                        )

                    return Result(
                        raw=Value[Signed[Result._width]](
                            self._val.msb(rest=cutoff).signed.resize(Result._width)
                            + do_round
                        )
                    )


#
#


class UFixed(Template[_FixedTemplateArg], AssignableType):
    _width: _FixedTemplateArg.width
    _exp: _FixedTemplateArg.exp

    @classmethod
    @pyeval
    def _count_bits_(cls):
        return cls._width

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier=Value):
        return cls(raw=bits.unsigned, _qualifier_=qualifier)

    def _to_bits_(self):
        return Value(self._val.bitvector)

    @classmethod
    @pyeval
    def min(cls):
        return 0

    @classmethod
    @pyeval
    def max(cls):
        return (2 ** (cls._width) - 1) * 2**cls._exp

    @classmethod
    @pyeval
    def _adjust_val(cls, val):
        return int(val / 2**cls._exp)

    @pyeval
    def __repr__(self):
        val = TypeQualifier.decay(self._val).to_int() * 2**self._exp
        return f"{type(self)}({val})"

    @classmethod
    @pyeval
    def right(cls):
        return cls._exp

    @classmethod
    @pyeval
    def left(cls):
        return cls._exp + cls._width - 1

    def __init__(self, val=None, *, raw=None, _qualifier_=Value):
        raw_type = Unsigned[self._width]

        if raw is not None:
            assert val is None, "val and raw cannot be set at the same time"
            assert instance_check(raw, raw_type), "raw type must match underlying type"
            self._val = _qualifier_(raw)
        else:
            if val is None or val is Full or val is Null:
                self._val = _qualifier_[raw_type](val)
            elif isinstance(val, (int, float)):
                static_assert(
                    self.min() <= val <= self.max(),
                    "value outside valid range of fixed point number",
                )
                self._val = _qualifier_[raw_type](self._adjust_val(val))
            elif instance_check(val, Unsigned):
                zeros = -self._exp
                static_assert(zeros >= 0)

                self._val = _qualifier_[raw_type](val.resize(self._width, zeros=zeros))
            elif instance_check(val, UFixed):
                if self.left() == val.left() and self.right() == val.right():
                    self._val = _qualifier_[raw_type](val._val)
                else:
                    assert self.left() >= val.left()
                    assert self.right() <= val.right()

                    zeros = self.right() - val.right()

                    self._val = _qualifier_[raw_type](
                        val._val.resize(self._width, zeros=zeros)
                    )
            else:
                raise AssertionError("invalid arg")

    #
    #
    #

    def _assign_(self, source, mode: AssignMode) -> None:
        if isinstance(source, UFixed):
            assert (
                source._width == self._width and source._exp == self._exp
            ), "source type ({}) does not match target type ({})".format(
                type(source), type(self)
            )
            self._val._assign_(source._val, mode)
        else:
            self._assign_(type(self)(source), mode)

    #
    #

    def __eq__(self, other: int | float | UFixed):
        if isinstance(other, (int, float)):
            return type(self)(other) == self
        else:
            assert isinstance(other, UFixed)
            assert type(other) is type(self)

            return self._val == other._val

    #
    #

    def __bool__(self):
        return bool(self._val)

    def __abs__(self):
        return type(self)(raw=Value(self._val))

    def __add__(self, other: UFixed):
        assert isinstance(other, UFixed)

        target_right = min(self.right(), other.right())
        target_left = max(self.left(), other.left()) + 1
        target_width = target_left - target_right + 1

        lhs_zeros = self.right() - target_right
        rhs_zeros = other.right() - target_right

        return UFixed[target_left:target_right](
            raw=self._val.resize(target_width, zeros=lhs_zeros)
            + other._val.resize(target_width, zeros=rhs_zeros)
        )

    def __sub__(self, other: UFixed):
        assert isinstance(other, UFixed)

        target_right = min(self.right(), other.right())
        target_left = max(self.left(), other.left()) + 1
        target_width = target_left - target_right + 1

        lhs_zeros = self.right() - target_right
        rhs_zeros = other.right() - target_right

        return UFixed[target_left:target_right](
            raw=self._val.resize(target_width, zeros=lhs_zeros)
            - other._val.resize(target_width, zeros=rhs_zeros)
        )

    def __mul__(self, other: UFixed):
        assert isinstance(other, UFixed)

        return UFixed[self.left() + other.left() + 1 : self.right() + other.right()](
            raw=self._val * other._val
        )

    resize = _Resize()

    def resize_fn(
        self,
        left: int,
        right: int,
        round_style: FixedRoundStyle = FixedRoundStyle.TRUNCATE,
        overflow_style: FixedRoundStyle = FixedOverflowStyle.WRAP,
    ):
        selfleft = self.left()
        selfright = self.right()

        if selfleft == left and selfright == right:
            return type(self)(raw=Value(self._val))

        Result = UFixed[left:right]

        if selfleft > left:
            overflow = selfleft - left

            if overflow_style is FixedOverflowStyle.WRAP:
                if selfright >= right:
                    zeros = selfright - right

                    if overflow >= self._val.width:
                        return Result(raw=Unsigned[Result._width](Null))
                    else:
                        return Result(
                            raw=self._val.lsb(rest=overflow).unsigned.resize(
                                zeros=zeros
                            )
                        )
                else:
                    cutoff = right - selfright

                    if round_style is FixedRoundStyle.TRUNCATE:
                        return Result(
                            raw=Value[Unsigned[Result._width]](
                                self._val.lsb(rest=overflow).msb(rest=cutoff).unsigned
                            )
                        )
                    else:
                        assert (
                            round_style is FixedRoundStyle.ROUND
                        ), "invalid round_style {}".format(round_style)

                        if cutoff == 1:
                            do_round = (
                                Unsigned[1](1)
                                if self._val[cutoff - 1] and self._val[cutoff]
                                else Unsigned[1](0)
                            )
                        else:
                            do_round = (
                                Unsigned[1](1)
                                if self._val[cutoff - 1]
                                and (self._val[cutoff] or self._val[cutoff - 2 : 0])
                                else Unsigned[1](0)
                            )

                        return Result(
                            raw=Value[Unsigned[Result._width]](
                                self._val.lsb(rest=overflow).msb(rest=cutoff).unsigned
                                + do_round
                            )
                        )
            else:
                assert (
                    overflow_style is FixedOverflowStyle.SATURATE
                ), "invalid overflow_style {}".format(overflow_style)
                overflow_bits = self._val.msb(selfleft - left)

                does_overflow = bool(overflow_bits)

                if selfright >= right:
                    zeros = selfright - right

                    if self._val.width <= overflow:
                        default_bits = Unsigned[Result._width](0)
                    else:
                        default_bits = self._val.lsb(rest=overflow).unsigned.resize(
                            Result._width, zeros=zeros
                        )

                    return Result(
                        raw=Value[Unsigned[Result._width]](
                            choose_first(
                                (does_overflow, Unsigned[Result._width].max()),
                                default=default_bits,
                            )
                        )
                    )
                else:
                    cutoff = right - selfright

                    if round_style is FixedRoundStyle.TRUNCATE:
                        return Result(
                            raw=Value[Unsigned[Result._width]](
                                choose_first(
                                    (does_overflow, Unsigned[Result._width].max()),
                                    default=self._val.lsb(rest=overflow)
                                    .msb(rest=cutoff)
                                    .unsigned,
                                )
                            )
                        )
                    else:
                        assert (
                            round_style is FixedRoundStyle.ROUND
                        ), "invalid round_style {}".format(round_style)

                        if cutoff == 1:
                            do_round = (
                                Unsigned[1](1)
                                if self._val[cutoff - 1] and self._val[cutoff]
                                else Unsigned[1](0)
                            )
                        else:
                            do_round = (
                                Unsigned[1](1)
                                if self._val[cutoff - 1]
                                and (self._val[cutoff] or self._val[cutoff - 2 : 0])
                                else Unsigned[1](0)
                            )

                        selected_bits = self._val.lsb(rest=overflow).msb(rest=cutoff)
                        overflow_or_full = does_overflow or not ~selected_bits

                        return Result(
                            raw=Value[Unsigned[Result._width]](
                                choose_first(
                                    (overflow_or_full, Unsigned[Result._width].max()),
                                    default=selected_bits.unsigned + do_round,
                                )
                            )
                        )

        else:
            if selfright >= right:
                return Result(
                    raw=self._val.resize(Result._width, zeros=selfright - right)
                )
            else:
                cutoff = right - selfright

                if round_style is FixedRoundStyle.TRUNCATE:
                    return Result(
                        raw=Value[Unsigned[Result._width]](
                            self._val.msb(rest=cutoff).unsigned.resize(Result._width)
                        )
                    )
                else:
                    assert (
                        round_style is FixedRoundStyle.ROUND
                    ), "invalid round_style {}".format(round_style)

                    if cutoff == 1:
                        do_round = (
                            Unsigned[1](1)
                            if self._val[cutoff - 1] and self._val[cutoff]
                            else Unsigned[1](0)
                        )
                    else:
                        do_round = (
                            Unsigned[1](1)
                            if self._val[cutoff - 1]
                            and (self._val[cutoff] or self._val[cutoff - 2 : 0])
                            else Unsigned[1](0)
                        )

                    return Result(
                        raw=Value[Unsigned[Result._width]](
                            self._val.msb(rest=cutoff).unsigned.resize(Result._width)
                            + do_round
                        )
                    )



from __future__ import annotations

from cohdl import std
from cohdl.std import Template
from cohdl.std import Record
from cohdl import (
    Bit,
    BitVector,
    Signal,
    Unsigned,
    Signed,
    Variable,
    Port,
    Entity,
    Null,
    signal,
    variable,
)

from functools import wraps

from typing import Generic, TypeVar, Union, overload

B = TypeVar("B")
T = TypeVar("T")
U = TypeVar("U")


class Bar:
    def bar(self) -> float:
        return 0.0


# Signal[Bit]   -> Signal[Bit]
# Signal[Bar]   -> Bar


class Foo:
    def foo(self):
        ...


class First(Record):
    x: Bit
    y: Bit
    z: Bit


class Rec(Record):
    a: Bit
    b: BitVector[9]

    f: First

    def __add__(self, other):
        return std.make(Rec, self.a ^ other.a, self.b & other.b, self.f)

    def __sub__(self, other):
        return Rec(self.a ^ other.a, self.b & other.b, self.f)


std.ref


class MyEntity(Entity):
    port_a = Port.input(Bit)
    port_b = Port.input(BitVector[9])

    port_x = Port.output(Bit)
    port_y = Port.output(BitVector[9])

    def architecture(self):
        my_array = std.StdArray[Rec, 100]()

        @std.concurrent
        def logic():
            s_first = Signal[First](self.port_a, self.port_a, self.port_a)

            ref_first = ...

            qwe = Variable[First]()

            # self.port_x <<= s_first.x

            # my_first = First(self.port_a, self.port_a, self.port_a)
            # my_rec = Rec(self.port_b[0], self.port_b, my_first)

            # my_first = std.make_ref(First, self.port_a, self.port_a, self.port_a)
            # my_first = First(self.port_a, self.port_a, self.port_a, _qualifier_=std.ref)
            # my_rec = std.make[std.ref](Rec, self.port_b[0], self.port_b, my_first)

            # print(my_first)
            # print(my_rec)

            # my_array[4] <<= my_rec
            # my_array[4].ref.f <<= my_first

            # my_array[4].ref.f <<= my_first
            # my_array[21] <<= my_rec

            ...

        def logic():
            f = First(self.port_a, self.port_a, self.port_a)

            rec_in = Rec(self.port_a, self.port_b, f)

            std.comment("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA")

            serialized = std.to_bits(rec_in)

            std.comment("AAAAAAAAAA")

            ser = std.make[Signal](std.Serialized[Rec], Null)

            ser.ref.b <<= rec_in.b
            ser.ref.f <<= rec_in.f

            std.comment("AAAAA")

            std.comment("BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB")

            deserialized = std.from_bits[Rec](serialized)

            std.comment("CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC")

            fo = First(self.port_x, self.port_x, self.port_x)

            rec_out = Rec(self.port_x, self.port_y, fo)

            std.comment("DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD")

            rec_out <<= deserialized - rec_in


print(std.VhdlCompiler.to_string(MyEntity))
std.make
exit()


class TemplateArg:
    def __init__(self, args):
        self.a, self.b = args

    def __hash__(self):
        return hash(self.a) ^ hash(self.b)

    def __eq__(self, other):
        return self.a == other.a and self.b == other.b

    def __str__(self):
        return f"{self.a},{self.b}"


class MyRecord(Record[TemplateArg]):
    a: BitVector[5]
    b: Bit
    c: Signed[7]
    d: Unsigned[TemplateArg.a]

    def __add__(self, other):
        return std.make(type(self), self.a, other.b, self.c + other.c, self.d - other.d)


m = MyRecord[4, 5]()
s = std.make[Signal](MyRecord[8, 9])
v = std.make[Variable](MyRecord[8, 9])

MyRecord()

print(m)
print(s)
print(v)
print(s + v)

print(std.to_bits(v))

exit()


class TemplateArg:
    def __init__(self, args):
        self.a, self.b = args

    def __hash__(self):
        return hash(self.a) ^ hash(self.b)

    def __eq__(self, other):
        return self.a == other.a and self.b == other.b

    def __str__(self):
        return f"{self.a},{self.b}"


class MyTemplate(Template[TemplateArg]):
    a: TemplateArg.a
    b: TemplateArg.b


Template_1_2 = MyTemplate[1, 2]
Template_4_2 = MyTemplate[4, 2]

print(Template_1_2.mro())

a = Template_1_2()
b = Template_4_2()

print(
    Template_1_2, a._template_meta_.annotations["a"], a._template_meta_.annotations["b"]
)
print(
    Template_4_2, b._template_meta_.annotations["a"], b._template_meta_.annotations["b"]
)


exit()

from cohdl import Signed, Unsigned, std, Signal, BitVector, Bit, Null


class MyRecord(std.Record):
    a: Unsigned[7]
    b: Signed[4]
    c: BitVector[43]
    d: Bit


class Derived(MyRecord):
    asdf: Bit


a = std.make(Derived)
b = std.make[Signal](Derived, a)
c = std.make[std.ref](Derived, a)

print()
print(repr(a))
print(std.to_bits(a))

print()
print(repr(b))
print(std.to_bits(b))

print()
print(repr(c))
print(std.to_bits(c))


print("\n\nDONE\n\n")
exit()


class WidthType(int):
    pass


class MyRecordTemplate(std.Record[WidthType]):
    a: Unsigned[WidthType]
    b: Signed[WidthType]


print("\n\nDONE\n\n")

m = MyRecordTemplate[4](Signal[Unsigned[4]](), b=Signed[4]())
print("DONE")
m = MyRecordTemplate[4](a=Signal[Unsigned[4]](), b=Signed[4]())
m = MyRecordTemplate[5](b=Signed[5](12), a=Unsigned[5](7))

m = std.make(MyRecordTemplate[3], Unsigned[3](7), Signed[3](1))
v = std.make(MyRecordTemplate[3], Null, ~Null)

# e = Example[123]()

bits = std.to_bits(m)

print(bits)

n = std.from_bits[MyRecordTemplate[3]](bits)

print(n)
print(v)
print(repr(n))
print(n == n)
print(n != n)



from __future__ import annotations

from cohdl._core import BitVector, Unsigned, AssignMode, Signal
from cohdl._core._intrinsic import _intrinsic
from ._template import Template
from ._core_utility import Value, to_bits, from_bits, count_bits
from ._assignable_type import AssignableType

import inspect


class _TemplateArg:
    def __init__(self, arg):
        if isinstance(arg, tuple):
            underlying, owner = arg
        else:
            underlying = arg
            owner = None

        if underlying is not None:
            assert isinstance(underlying, type)

        self.underlying = underlying
        self.owner = owner

    def __hash__(self) -> int:
        return hash((self.underlying, self.owner))

    def __eq__(self, other: _TemplateArg):
        return self.underlying is other.underlying and self.owner is other.owner


class Enum(Template[_TemplateArg], AssignableType):
    _underlying_: _TemplateArg.underlying
    _default_: Enum
    __members__: dict[str, Enum]

    @property
    def raw(self):
        return self._val

    @raw.setter
    def raw(self, value):
        assert (
            value is self._val
        ), "assignment to the raw property only allowed using `<<=`, `^=` or `@=`"
        pass

    @property
    @_intrinsic
    def name(self):
        assert hasattr(
            self, "_name"
        ), "the name property is only defined for the constant enumerators used in the declaration"
        return self._name

    @property
    @_intrinsic
    def info(self):
        assert hasattr(
            self, "_info"
        ), "the info property is only defined for the constant enumerators used in the declaration"
        return self._info

    @classmethod
    @_intrinsic
    def _count_bits_(cls):
        return count_bits(cls._underlying_)

    @classmethod
    def _from_bits_(cls, bits, qualifier=Value):
        return cls(
            from_bits[cls._underlying_](bits, qualifier),
            _qualifier_=qualifier,
            _unsafe_init=True,
        )

    def _to_bits_(self):
        return to_bits(self._val)

    @classmethod
    def _unsafe_init_(cls, value, _qualifier_=Value):
        return cls(value, _qualifier_=_qualifier_, _unsafe_init=True)

    def _assign_(self, source, mode: AssignMode) -> None:
        assert type(self) is type(source)
        self._val._assign_(source._val, mode)

    @classmethod
    def _template_deduce_(cls, *args, **kwargs) -> Template[_TemplateArg]:
        return Enum[None]

    def _init_enumerator(self, val, info: str | None = None):
        self._val = val
        self._info = info

    def _init_normal(self, val=None, _qualifier_=Value, _unsafe_init=False):
        underlying = type(self)._underlying_

        if val is None:
            self._val = _qualifier_[underlying](self._default_._val)
        elif isinstance(val, Enum):
            if type(val) is type(self):
                self._val = _qualifier_[underlying](val._val)
            else:
                assert val._underlying_ is None
                assert _unsafe_init, "invalid argument type"
                self._val = _qualifier_[underlying](val._val)
                self._info = val._info
                self._name = val._name
        else:
            assert _unsafe_init, "invalid argument type"
            self._val = _qualifier_[underlying](val)

    def __init__(self, *args, **kwargs):

        if type(self)._underlying_ is None:
            self._init_enumerator(*args, **kwargs)
        else:
            self._init_normal(*args, **kwargs)

    def __init_subclass__(cls) -> None:
        if not hasattr(cls, "_underlying_"):
            # template not yet specialized
            return

        if "_underlying_" in cls.__dict__:
            # template specialized but not used class has been derived
            return

        enumerators = {}

        for name, value in cls.__dict__.items():
            if name.startswith("__") or inspect.isfunction(value):
                continue

            if isinstance(value, Enum):
                enumerator = value
            else:
                if isinstance(value, tuple):
                    value, info = value
                else:
                    info = None

                enumerator = Enum(value, info)

            enumerator._name = name
            enumerators[name] = cls(enumerator, _unsafe_init=True)

        is_first = True

        for name, enumerator in enumerators.items():
            if is_first:
                is_first = False
                setattr(cls, "_default_", enumerator)
                setattr(cls, "__members__", enumerators)
            setattr(cls, name, enumerator)

    def __eq__(self, other):
        assert type(self) is type(other)
        return self._val == other._val

    def __ne__(self, other):
        assert type(self) is type(other)
        return self._val != other._val

    def __bool__(self):
        return bool(self._val)


class FlagEnum(Enum):
    def __and__(self, other):
        assert type(self) is type(other)
        return type(self)(self._val & other._val, _unsafe_init=True)

    def __or__(self, other):
        assert type(self) is type(other)
        return type(self)(self._val | other._val, _unsafe_init=True)

    def __xor__(self, other):
        assert type(self) is type(other)
        return type(self)(self._val ^ other._val, _unsafe_init=True)



from __future__ import annotations


class MakeTarget:
    def __init__(
        self,
        target_str,
        commands: list[str] | str,
        dep: list[MakeTarget | str] = [],
        required_by: list[MakeTarget | str] = [],
        *,
        order_only: list[MakeTarget | str] = [],
        phony=False,
        lock=False,
    ):
        assert not (phony and lock), "lock cannot be set for phony target"

        if isinstance(commands, str):
            commands = [commands]

        self.target_str = target_str
        self.commands = commands
        self.dep: dict[str, str | MakeTarget] = {}
        self.order_only: dict[str, str | MakeTarget] = {}

        self.add_dependency(*dep)
        self.add_dependency(*order_only, order_only=True)

        self.required_by: list[str] = [
            x if isinstance(x, str) else x.target_str for x in required_by
        ]

        self.phony = phony
        self.lock = lock

    def add_dependency(self, *dep: str | MakeTarget, order_only=False):
        if order_only:
            result_dict = self.order_only
        else:
            result_dict = self.dep

        for d in dep:
            d_str = d if isinstance(d, str) else d.target_str
            result_dict[d_str] = d

    def __hash__(self) -> int:
        return id(self)

    def collect_all_targets(self) -> dict[str, MakeTarget]:
        result: dict[str, MakeTarget] = {self.target_str: self}

        for dep in self.dep.values():
            if isinstance(dep, MakeTarget):
                result = {**result, **dep.collect_all_targets()}

        for dep in self.order_only.values():
            if isinstance(dep, MakeTarget):
                result = {**result, **dep.collect_all_targets()}

        return result

    def generate_makefile(*targets: MakeTarget, path: str, lock_dir: str | None = None):
        all_targets: dict[str, MakeTarget] = {}

        if lock_dir is not None:
            lock_dir_target = MakeTarget(lock_dir, f"mkdir {lock_dir}")
        else:
            lock_dir_target = None

        for target in targets:
            all_targets = {**all_targets, **target.collect_all_targets()}

        for prerequisit in all_targets.values():
            for r in prerequisit.required_by:
                all_targets[r].add_dependency(prerequisit)

        with open(path, "w+") as file:
            phony = [
                target.target_str for target in all_targets.values() if target.phony
            ]

            lock_targets = [
                target.target_str for target in all_targets.values() if target.lock
            ]

            if len(phony) != 0:
                file.writelines([f".PHONY: {' '.join(phony)}\n", "\n"])

            def write_target(target: MakeTarget):
                cmds = [*target.commands]

                if target.lock:
                    assert (
                        lock_dir_target is not None
                    ), "target.lock is set but no lock_dir was specified"
                    target_str = f"{lock_dir}/{target.target_str}"
                    target.add_dependency(lock_dir_target, order_only=True)
                    cmds.append(f"touch {target_str}")
                else:
                    target_str = target.target_str

                def format_dep(dep: str):
                    if dep in lock_targets:
                        return f"{lock_dir}/{dep}"
                    return dep

                normal_dep = " ".join([format_dep(dep) for dep in target.dep.keys()])
                order_only = [format_dep(dep) for dep in target.order_only.keys()]

                if len(order_only) == 0:
                    order_only_dep = ""
                else:
                    order_only_dep = " | " + " ".join(order_only)

                file.writelines(
                    [
                        f"{target_str}: {normal_dep}{order_only_dep}\n",
                        *[f"\t{cmd}\n" for cmd in cmds],
                        "\n",
                    ]
                )

            # write targets received as arguments in order (ensure, that first target appears first in Makefile)
            for target in targets:
                write_target(target)
                del all_targets[target.target_str]

            # write remaining targets
            for target in all_targets.values():
                write_target(target)

            if lock_dir_target is not None:
                write_target(lock_dir_target)



from __future__ import annotations

import inspect
import typing

from typing import Any

from cohdl._core._type_qualifier import (
    TypeQualifierBase,
    TypeQualifier,
    Temporary,
    Signal,
    Variable,
)

from cohdl._core import (
    Bit,
    BitVector,
    Signed,
    Unsigned,
    select_with,
    evaluated,
    Null,
    Full,
    static_assert,
    is_primitive_type,
)

from cohdl._core._intrinsic import _intrinsic

from cohdl._core._intrinsic import comment as cohdl_comment


def nop(*args, **kwargs):
    pass


def comment(*lines):
    cohdl_comment(*lines)


@_intrinsic
def as_pyeval(fn, /, *args, **kwargs):
    return fn(*args, **kwargs)


@_intrinsic
def fail(message: str = "", *args, **kwargs):
    raise AssertionError("Compilation failed: " + message.format(*args, **kwargs))


@_intrinsic
def identity(inp, /):
    return inp


@_intrinsic
def _bind_partial(signature, args, kwargs) -> inspect.Signature:
    return signature.bind_partial(*args, **kwargs)


def _determine_defaults(signature: inspect.Signature, args, kwargs, scope):
    bound = _bind_partial(signature, args, kwargs)

    for name, param in as_pyeval(signature.parameters.items):
        if name not in bound.arguments:
            param_str = str(param)

            if not as_pyeval(param_str.startswith, "*"):
                _, default_str = as_pyeval(param_str.split, "=", maxsplit=1)

                default = as_pyeval(eval, default_str, scope)
                as_pyeval(bound.arguments.__setitem__, name, default)

    return bound


def regenerate_defaults(fn, /):
    assert (
        not evaluated()
    ), "regenerate_defaults should not be called in synthesizable contexts"

    is_function = inspect.isfunction(fn)
    is_async = inspect.iscoroutinefunction(fn)

    assert (
        is_function or is_async
    ), "the object wrapped by regnerate_defaults must be a function or async function"

    scope = fn.__globals__
    callable = fn
    signature = inspect.signature(callable)

    if not is_async:

        def helper(*args, **kwargs):
            bound = _determine_defaults(signature, args, kwargs, scope)
            return fn(*bound.args, **bound.kwargs)

    else:

        async def helper(*args, **kwargs):
            bound = _determine_defaults(signature, args, kwargs, scope)
            return fn(*bound.args, **bound.kwargs)

    return helper


class _Value:
    def __init__(self, T=None):
        self._T = T

    def _make_const_or_temp(self, T, arg):
        if isinstance(arg, TypeQualifier):
            return Temporary[T](arg)
        elif isinstance(arg, TypeQualifierBase):
            return TypeQualifierBase.decay(arg)
        else:
            return T(arg)

    def __call__(self, *args, **kwargs):
        if self._T is None:
            assert len(args) == 1 and len(kwargs) == 0
            T = type(TypeQualifierBase.decay(args[0]))
        else:
            T = self._T

        if is_primitive_type(T):
            if len(args) == 0 and len(kwargs) == 0:
                return T()

            assert len(args) == 1 and (
                len(kwargs) == 0 or (len(kwargs) == 1 and kwargs.__contains__("name"))
            )
            arg = args[0]

            if isinstance(arg, TypeQualifier):
                if isinstance(arg, Temporary[T]):
                    return arg
                else:
                    return Temporary[T](arg, **kwargs)
            elif isinstance(arg, TypeQualifierBase):
                return TypeQualifierBase.decay(arg)
            else:
                return T(arg, **kwargs)
        else:
            return T(*args, **kwargs, _qualifier_=Value)

    def __getitem__(self, arg):
        assert isinstance(arg, type)

        return _Value(arg)


class _Ref:
    def __init__(self, T=None):
        self._T = T

    @_intrinsic
    def __call__(self, *args, **kwargs):
        if self._T is None:
            assert len(args) == 1 and len(kwargs) == 0
            T = type(TypeQualifierBase.decay(args[0]))
        else:
            T = self._T

        if is_primitive_type(T):
            assert len(args) == 1 and (
                len(kwargs) == 0 or (len(kwargs) == 1 and kwargs.__contains__("name"))
            )
            arg = args[0]

            if subclass_check(T, BitVector):
                if subclass_check(T, Signed):
                    return arg.signed
                elif subclass_check(T, Unsigned):
                    return arg.unsigned
                else:
                    return arg.bitvector
            else:
                assert instance_check(arg, T)
                return arg
        else:
            return T(*args, **kwargs, _qualifier_=Ref)

    def __getitem__(self, arg):
        return _Ref(arg)


#
#
#


class _Nonlocal:
    def __init__(self, qualified_type):
        self._qualified_type = qualified_type

    @_intrinsic
    def __call__(self, *args, **kwargs):
        assert (
            self._qualified_type is not None
        ), "type of Nonlocal must be set using Nonlocal[TARGET_TYPE]"
        return self._qualified_type(*args, **kwargs)

    @_intrinsic
    def __getitem__(self, arg):
        if self._qualified_type is None:
            assert issubclass(
                arg, (Signal, Variable)
            ), "std.Nonlocal should only be used to create Signals or Variables"
            return _Nonlocal(arg)
        else:
            return _Nonlocal(self._qualified_type[arg])


class _Noreset:
    _Qualifier: Signal

    def __init__(self, T=None):
        self._T = T

    @_intrinsic
    def __call__(self, *args, **kwargs):
        if self._T is None:
            assert len(args) == 1 and len(kwargs) == 0
            T = type(TypeQualifierBase.decay(args[0]))
        else:
            T = self._T

        if is_primitive_type(T) or T is bool or T is int:
            return self._Qualifier[T](*args, **kwargs, noreset=True)
        else:
            return T(*args, **kwargs, _qualifier_=type(self)())

    def __getitem__(self, arg):
        return type(self)(arg)


class _NoresetSignal(_Noreset):
    _Qualifier = Signal


class _NoresetVariable(_Noreset):
    _Qualifier = Variable


Ref = _Ref()
Value = _Value()
Nonlocal = _Nonlocal(None)
NoresetSignal = _NoresetSignal()
NoresetVariable = _NoresetVariable()

#
#
#


@_intrinsic
def iscouroutinefunction(fn):
    return inspect.iscoroutinefunction(fn)


def base_type(x):
    if isinstance(x, type):
        if issubclass(x, TypeQualifierBase):
            return x.type
        else:
            return x
    else:
        if isinstance(x, TypeQualifierBase):
            return x.type
        else:
            return type(x)


def instance_check(val, type):
    return isinstance(TypeQualifierBase.decay(val), type)


def subclass_check(val, type):
    return issubclass(TypeQualifierBase.decay(val), type)


async def as_awaitable(fn, /, *args, **kwargs):
    if iscouroutinefunction(fn):
        return await fn(*args, **kwargs)
    else:
        return fn(*args, **kwargs)


#
#
#


@_intrinsic
def zeros(len: int):
    return BitVector[len](Null)


@_intrinsic
def ones(len: int):
    return BitVector[len](Full)


@_intrinsic
def width(inp: Bit | BitVector) -> int:
    if instance_check(inp, Bit):
        return 1
    else:
        return inp.width


def one_hot(width: int, bit_pos: int | Unsigned) -> BitVector:
    assert 0 <= bit_pos < width, "bit_pos out of range"
    return (Unsigned[width](1) << bit_pos).bitvector


def reverse_bits(inp: BitVector) -> BitVector:
    return concat(*inp)


def is_qualified(arg):
    return isinstance(arg, TypeQualifierBase)


def const_cond(arg):
    result = bool(arg)
    static_assert(isinstance(result, bool), "condition is not a constant")
    return result


class _UncheckedType:
    pass


def _check_type(result, expected):
    if expected is _UncheckedType:
        return True
    elif expected is None:
        return result is None
    else:
        return instance_check(result, expected)


@_intrinsic
def _format_type_check_error(expected_type, actual_type):
    return f"invalid type in checked expression: expected '{expected_type}' but got '{actual_type}'"


class _TypeCheckedExpression:
    def __init__(self, expected_type: type | tuple = _UncheckedType):
        self._expected = expected_type

    def _checked(self, arg):
        assert _check_type(arg, self._expected), _format_type_check_error(
            self._expected, arg
        )
        return arg

    def __getitem__(self, expected_type):
        return type(self)(expected_type)


class _CheckType(_TypeCheckedExpression):
    def __call__(self, arg):
        assert (
            self._expected is not _UncheckedType
        ), "no type specified in std.check_type"
        return self._checked(arg)


class _Select(_TypeCheckedExpression):
    def __call__(self, arg, branches: dict, default=None):
        return self._checked(
            select_with(
                arg,
                branches,
                default=default,
            )
        )


def _first_impl(*args, default):
    if len(args) == 0:
        return default
    else:
        first, *rest = args
        return first[1] if first[0] else _first_impl(*rest, default=default)


class _ChooseFirst(_TypeCheckedExpression):
    def __call__(self, *args, default):
        return self._checked(_first_impl(*args, default=default))


class _Cond(_TypeCheckedExpression):
    def __call__(self, cond: bool, on_true, on_false):
        return self._checked(on_true if cond else on_false)


check_type = _CheckType()
select = _Select()
choose_first = _ChooseFirst()
cond = _Cond()

#
#
#


@_intrinsic
def count_bits(inp):
    if not isinstance(inp, type):
        inp = type(inp)

    if subclass_check(inp, Bit):
        return 1
    elif subclass_check(inp, BitVector):
        return inp.width
    else:
        return inp._count_bits_()


def to_bits(inp, /):
    if instance_check(inp, (Bit, BitVector)):
        return as_bitvector(inp)
    else:
        return inp._to_bits_()


class _FromBits:
    @_intrinsic
    def __init__(self, target_type: type):
        self._target_type = target_type

        if target_type is not None:
            assert isinstance(target_type, type), "target_type is not a type"
            assert issubclass(target_type, (Bit, BitVector)) or hasattr(
                target_type, "_from_bits_"
            ), "target_type does not implement from_bits"

    @_intrinsic
    def __getitem__(self, target_type):
        return _FromBits(target_type)

    def __call__(self, bits: BitVector, qualifier=Value):
        assert (
            self._target_type is not None
        ), "missing target type, expected usage: std.from_bits[TARGET_TYPE](inp)"
        assert instance_check(
            bits, BitVector
        ), "std.from_bits expects a BitVector as its first argument"
        bv = bits.bitvector

        if issubclass(self._target_type, Bit):
            assert bv.width == 1, "cannot deserialize BitVector to Bit, width is not 1"
            return qualifier[Bit](bv[0])
        elif issubclass(self._target_type, BitVector):
            assert (
                bv.width == self._target_type.width
            ), "cannot deserialize BitVector to BitVector, width does not match"
            return qualifier[self._target_type](bv)
        else:
            assert (
                self._target_type._count_bits_() == bits.width
            ), "width of serialized BitVector does not match width expected by the target type"
            return self._target_type._from_bits_(bits, qualifier)


from_bits = _FromBits(None)

#
#
#


@_intrinsic
def _get_return_type_hint(fn):
    return typing.get_type_hints(fn).get("return", _UncheckedType)


def check_return(fn):
    if iscouroutinefunction(fn):

        async def wrapper(*args, **kwargs):
            result = await fn(*args, **kwargs)
            assert _check_type(
                result, _get_return_type_hint(fn)
            ), "invalid return value in function call"
            return result

    else:

        def wrapper(*args, **kwargs):
            result = fn(*args, **kwargs)
            assert _check_type(
                result, _get_return_type_hint(fn)
            ), "invalid return value in function call"
            return result

    wrapper.__name__ = fn.__name__
    return wrapper


#
#
#


def binary_fold(fn, args, right_fold=False):
    if len(args) == 1:
        return Value(args[0])
    else:
        if const_cond(right_fold):
            first, *rest = args
            return fn(first, binary_fold(fn, rest, right_fold=True))
        else:
            first, snd, *rest = args
            return binary_fold(fn, [fn(first, snd), *rest])


@_intrinsic
def _batch_args(args: list, batch_size: int):
    batches = []

    for nr in range(0, len(args), batch_size):
        batches.append(args[nr : nr + batch_size])

    return batches


def batched_fold(fn, args, batch_size=2):
    if const_cond(len(args) <= batch_size):
        return binary_fold(fn, *args)
    else:
        return batched_fold(
            fn,
            *[
                batched_fold(fn, batch, batch_size=batch_size)
                for batch in _batch_args(args, batch_size)
            ],
        )


def _concat_pairwise(args: list):
    result = []

    for a, b in zip(args[0::2], args[1::2]):
        as_pyeval(result.append, a @ b)

    return result


def concat(first, *args):
    if len(args) == 0:
        if instance_check(first, Bit):
            return as_bitvector(first)
        else:
            assert instance_check(
                first, BitVector
            ), "first is {} and not a BitVector".format(first)
            return Value[BitVector[len(first)]](first)
    else:
        if len(args) % 2 == 0:
            return first @ concat(*_concat_pairwise(args))
        else:
            return concat(*_concat_pairwise([first, *args]))


def _repeat_filter_by_factor(stretch_list, factor: int):
    return [s for nr, s in enumerate(stretch_list) if (factor & (1 << nr))]


def repeat(val, times: int):
    pow2_list = [val]

    for _ in range(1, times.bit_length()):
        as_pyeval(pow2_list.append, pow2_list[-1] @ pow2_list[-1])

    return concat(
        *[elem for elem in as_pyeval(_repeat_filter_by_factor, pow2_list, times)]
    )


def stretch(val: Bit | BitVector, factor: int):
    assert factor > 0

    if instance_check(val, Bit):
        return repeat(val, factor)
    elif instance_check(val, BitVector):
        if const_cond(factor == 1):
            return val.bitvector.copy()
        else:
            # reverse stretched list so bit 0 is the leftmost
            # concatenated value
            return concat(*[stretch(b, factor) for b in val][::-1])
    else:
        raise AssertionError("invalid argument")


def resize(inp: BitVector, result_width: int):
    inp_width = inp.width

    assert (
        inp_width <= result_width
    ), f"the result width of resize must be larger than the input width"

    if inp_width == result_width:
        return Value[BitVector[result_width]](inp)
    else:
        return zeros(inp_width) @ inp


def leftpad(inp: BitVector, result_width: int, fill=None):
    inp_width = inp.width

    assert (
        inp_width <= result_width
    ), f"the result width of resize must be larger than the input width"

    if inp_width == result_width:
        return Value[BitVector[result_width]](inp)
    else:
        if fill is None:
            fill_bit = Bit(False)
        elif fill is Null or fill is Full:
            fill_bit = Bit(fill)
        else:
            assert instance_check(
                fill, Bit
            ), f"parameter 'fill' should be Null, Full or a Bit type"
            fill_bit = fill

        return stretch(fill_bit, result_width - inp_width) @ inp


def rightpad(inp: BitVector, result_width: int, fill=None):
    inp_width = inp.width

    assert (
        inp_width <= result_width
    ), f"the result width of resize must be larger than the input width"

    if inp_width == result_width:
        return Value[BitVector[result_width]](inp)
    else:
        if fill is None:
            fill_bit = Bit(False)
        elif fill is Null or fill is Full:
            fill_bit = Bit(fill)
        else:
            assert instance_check(
                fill, Bit
            ), f"parameter 'fill' should be Null, Full or a Bit type"
            fill_bit = fill

        return inp @ stretch(fill_bit, result_width - inp_width)


def pad(
    inp: BitVector, left: int = 0, right: int = 0, fill: Bit | Null | Full = Null
) -> BitVector:
    if fill is None:
        fill_bit = Bit(False)
    elif fill is Null or fill is Full:
        fill_bit = Bit(fill)
    else:
        assert instance_check(
            fill, Bit
        ), f"parameter 'fill' should be Null, Full or a Bit type"
        fill_bit = fill

    assert isinstance(left, int) and left >= 0
    assert isinstance(right, int) and right >= 0

    if left == 0 and right == 0:
        return as_bitvector(inp)

    if left != 0:
        left_padded = stretch(fill_bit, left) @ inp
    else:
        left_padded = inp

    if right != 0:
        return left_padded @ stretch(fill_bit, right)
    else:
        return left_padded


def apply_mask(old: BitVector, new: BitVector, mask: BitVector):
    assert old.width == new.width, "old.width does not match new.width"
    return (old.bitvector & ~mask) | (new.bitvector & mask)


class Mask:
    def __init__(self, val):
        self._val = val

    def apply(self, old, new):
        if self._val is Null:
            return Temporary(old)
        elif self._val is Full:
            return Temporary(new.copy())
        else:
            return apply_mask(old, new, self._val)

    def as_vector(self, width: int):
        return self.apply(zeros(width), ones(width))


def as_bitvector(inp: BitVector | Bit | str):
    if isinstance(inp, str):
        return BitVector[len(inp)](inp)
    elif instance_check(inp, BitVector):
        return inp.bitvector.copy()
    else:
        assert instance_check(inp, Bit)
        return (inp @ inp)[0:0]


def rol(inp: BitVector, n: int = 1) -> BitVector:
    static_assert(0 <= n <= inp.width)
    if const_cond(n == 0 or n == inp.width):
        return inp.bitvector.copy()
    else:
        return inp.lsb(rest=n) @ inp.msb(n)


def ror(inp: BitVector, n: int = 1) -> BitVector:
    static_assert(0 <= n <= inp.width)
    if const_cond(n == 0 or n == inp.width):
        return inp.bitvector.copy()
    else:
        return inp.lsb(n) @ inp.msb(rest=n)


def lshift_fill(val: BitVector, fill: Bit | BitVector) -> BitVector:
    width_val = width(val)
    width_fill = width(fill)

    if width_fill == width_val:
        return as_bitvector(fill)
    elif width_fill > width_val:
        fail("fill width ({}) exceeds width of value ({})", width_fill, width_val)
    else:
        return val.lsb(width_val - width_fill) @ fill


def rshift_fill(val: BitVector, fill: Bit | BitVector) -> BitVector:
    width_val = width(val)
    width_fill = width(fill)

    if width_fill == width_val:
        return as_bitvector(fill)
    elif width_fill > width_val:
        fail("fill width ({}) exceeds width of value ({})", width_fill, width_val)
    else:
        return fill @ val.msb(width_val - width_fill)


def batched(input: BitVector, n: int) -> list[BitVector]:
    static_assert(len(input) % n == 0)

    return [input[off + n - 1 : off] for off in range(0, len(input), n)]


def select_batch(
    input: BitVector, onehot_selector: BitVector, batch_size: int
) -> BitVector:
    static_assert(len(input) == len(onehot_selector) * batch_size)
    masked = input.bitvector & stretch(onehot_selector, batch_size)

    return binary_fold(lambda a, b: a | b, batched(masked, batch_size))


def parity(vec: BitVector) -> Bit:
    return binary_fold(lambda a, b: a ^ b, vec)



from __future__ import annotations

from typing import TypeVar, Coroutine

def _intrinsic(fn):
    return fn

def comment(*lines: str) -> None:
    """
    Inserts a comment into the generated VHDL representation.

    std.comment("Hello, world!", "A", "B") is translated into

    `-- Hello, world!`

    `-- A`

    `-- B`
    """

Option = TypeVar("Option")
Result = TypeVar("Result")

def select_with(
    arg, branches: dict[Option, Result], default: Result | None = None
) -> Result:
    if arg in branches:
        return branches[arg]

    return default

#
#
#

def coroutine_step(coro: Coroutine) -> None:
    """
    This function is the core of cohdls coroutine support

    The cohdl compiler translates the given coroutine into
    a HDL statemachine and inserts it at the location of
    the call to `coroutine_step`.

    This has the effect, that the statemachine advances
    each time `coroutine_step` is "called" at runtime.
    """

class sensitivity:
    @staticmethod
    def list(*args) -> None:
        """
        this function is used in sequential contexts
        the compiler adds the arguments (which have to be
        signals) to the sensitivity list of the generated hdl process
        """
    @staticmethod
    def all() -> None:
        """
        this function is used in sequential contexts
        when the compiler encounters this function it
        adds all signals, that are read in the context
        to the sensitivity list of the generated hdl process
        """

#
#
#

def reset_context() -> None:
    """
    This function can only be used in sequential contexts.

    When the cohdl compiler encounters a call to `reset_context`
    it inserts default assignments for all Signals, Variables and Ports that
    have a defined default value and are driven somewhere in the enclosing context
    """

def reset_pushed() -> None:
    """
    `reset_pushed` is similar to `reset_context`. The only difference is,
    that only Ports/Signals using the push assignment mode (`^=` operator) are considered.

    Within a sequential context, that does not call `reset_pushed`
    push assignments are functionally equivalent to signal assignment.
    """

#
#
#

def rising_edge(_) -> bool: ...
def falling_edge(_) -> bool: ...
def high_level(_) -> bool: ...
def low_level(_) -> bool: ...



from __future__ import annotations

from cohdl import pyeval

from cohdl import enum

enum.Enum()


class _MetaClassInfo:
    def __init__(self, original_init):
        self._original_init = original_init


class MetaMember:
    def __class_getitem__(cls, args):
        class _MetaMember(cls):
            pass

        setattr(_MetaMember, "@meta_args", args)
        return _MetaMember

    def _meta_config_(self, parent, /, *args, **kwargs):
        pass


@pyeval
def _wrapped_init_(self: MetaObject, *args, **kwargs):
    meta_class_info: _MetaClassInfo = getattr("@meta_class_info", type(self))

    self._interface_ = MetaInterface()
    meta_class_info._original_init(self, *args, **kwargs)

    annotations = type(self).__annotations__

    assert isinstance(self._interface_, MetaInterface)

    for name, ann in annotations.items():
        if issubclass(ann, MetaMember):
            if name not in self._interface_.__dict__:
                setattr(self._interface_, name, ann._default_construct_())

            member = self._interface_.__dict__[name]

            assert isinstance(member, MetaMember)
            assert not member._reg_is_configured_()

            member._meta_config_(self, name, ann._meta_args_())

    for value in self._interface_.__dict__.values():
        assert isinstance(value, MetaMember)
        assert value._reg_is_configured()

    for name, value in self.__dict__.items():
        if isinstance(value, MetaMember):
            assert (
                name in self._interface_.__dict__
            ), "the object '{name}' is not part of the interface definition of this class"

    self._interface_._frozen = True


class MetaInterface:
    _frozen = False

    def __setattr__(self, attr, value):
        if getattr(self, "_frozen"):
            raise AttributeError("Trying to set attribute on a frozen instance")
        return super().__setattr__(attr, value)

    def _cohdlstd_on_freeze(self):
        for name, value in self.__dict__.items():
            assert isinstance(
                value, MetaMember
            ), f"'{name}' is not a valid interface member"

    def _cohdlstd_members(self) -> list[MetaMember]:
        return [*self.__dict__.values()]

    def _cohdlstd_sorted_members(self) -> list[MetaMember]:
        return sorted(self._cohdl_members_(), lambda m: m._offset_)


class MetaObject:
    _cohdlstd_interface_type: type[MetaInterface]

    _interface_: MetaInterface

    def __init_subclass__(cls) -> None:
        cls._cohdlstd_reg_class_info = _MetaClassInfo(cls.__init__)
        cls.__init__ = _wrapped_init_



from __future__ import annotations
from ._assignable_type import AssignableType
from cohdl._core._intrinsic import _intrinsic

import cohdl
import typing

from ._core_utility import Ref, instance_check, fail
from ._template import Template


class BitArg(int):
    pass


class FieldBit(Template[BitArg]):
    _offset_: BitArg

    @_intrinsic
    def __new__(cls, source: cohdl.BitVector):
        return source[cls._offset_]


class BitVectorArg:
    start: int
    stop: int
    type: type

    @_intrinsic
    def __init__(self, arg: slice) -> None:
        if not isinstance(arg, tuple):
            arg = (arg, cohdl.BitVector)

        span: slice = arg[0]

        assert isinstance(span, slice), "first generic argument must be a slice"
        assert isinstance(span.start, int), "start parameter must be an integer"
        assert isinstance(span.stop, int), "stop parameter must be an integer"
        assert span.step is None, "step parameter not allowed"
        assert arg[1] in (
            cohdl.BitVector,
            cohdl.Signed,
            cohdl.Unsigned,
        ), f"invalid argument type {arg[1]}"

        self.start = span.start
        self.stop = span.stop
        self.type = arg[1]

    @_intrinsic
    def __str__(self) -> str:
        return f"{self.start}:{self.stop},{self.type}"

    def __hash__(self) -> int:
        return hash((self.start, self.stop, self.type))

    def __eq__(self, other: BitVectorArg):
        return (
            self.start == other.start
            and self.stop == other.stop
            and self.type is other.type
        )


class FieldBitVector(Template[BitVectorArg]):
    _start_: BitVectorArg.start
    _stop_: BitVectorArg.stop
    _type_: BitVectorArg.type

    BitVector: FieldBitVector[BitVectorArg.start : BitVectorArg.stop, cohdl.BitVector]
    Signed: FieldBitVector[BitVectorArg.start : BitVectorArg.stop, cohdl.Signed]
    Unsigned: FieldBitVector[BitVectorArg.start : BitVectorArg.stop, cohdl.Unsigned]

    @_intrinsic
    def __new__(cls, source: cohdl.BitVector):
        if cls._type_ is cohdl.BitVector:
            return source[cls._start_ : cls._stop_]
        elif cls._type_ is cohdl.Unsigned:
            return source[cls._start_ : cls._stop_].unsigned
        elif cls._type_ is cohdl.Signed:
            return source[cls._start_ : cls._stop_].signed
        else:
            fail("invalid target type {}", cls._type_)


class Field:
    def __class_getitem__(cls, arg):
        if isinstance(arg, int):
            return FieldBit[arg]
        else:
            return FieldBitVector[arg]


_bitfield_classes: dict[int, type[BitField]] = {}


class BitField(AssignableType):
    @_intrinsic
    def __class_getitem__(cls, arg):
        assert isinstance(arg, int)

        if arg not in _bitfield_classes:
            newtype = type(
                f"BitField[{arg}]",
                (_BitFieldInst,),
                {
                    "_cohdlstd_is_bitfield_base": True,
                    "_width_": arg,
                    "_offset_": None,
                },
            )

            _bitfield_classes[arg] = newtype

        return _bitfield_classes[arg]

    def __init__(self, vec: cohdl.BitVector | None = None, *, _qualifier_=Ref):
        if vec is None:
            self._vec = _qualifier_[cohdl.BitVector[self._width_]]()
        elif vec is cohdl.Null or vec is cohdl.Full:
            self._vec = _qualifier_[cohdl.BitVector[self._width_]](vec)
        elif isinstance(vec, BitField):
            assert isinstance(
                vec, type(self)
            ), "bitfield type of argument does not match target"
            self._vec = _qualifier_(vec._vec)
        else:
            assert (
                vec.width == self._width_
            ), "vector width does not match bitfield width"
            self._vec = _qualifier_(vec)

        for name, Field in self._fields_.items():
            if issubclass(Field, (FieldBit, FieldBitVector)):
                setattr(self, name, Field(self._vec))
            else:
                assert issubclass(
                    Field, BitField
                ), "internal error: member of BitField has unexpected type {}".format(
                    Field
                )
                assert (
                    Field._offset_ is not None
                ), "the offset of sub-BitFields must be specified"

                subvec = self._vec.msb(rest=Field._offset_).lsb(Field._width_)
                setattr(self, name, Field(subvec, _qualifier_=Ref))

    def _assign_(self, source, mode: cohdl.AssignMode):
        if isinstance(source, BitField):
            assert isinstance(source, type(self))
            self._vec._assign_(source._vec, mode)
        elif isinstance(source, dict):
            for name, value in source.items():
                getattr(self, name)._assign_(value, mode)
        else:
            assert instance_check(source, cohdl.BitVector[self._width_])
            self._vec._assign_(source, mode)

    @classmethod
    @_intrinsic
    def _count_bits_(cls):
        return cls._width_

    @classmethod
    def _from_bits_(cls, bitvector: cohdl.BitVector, qualifier):
        return cls(bitvector, _qualifier_=qualifier)

    def _to_bits_(self):
        return cohdl.Temporary(self._vec)

    @_intrinsic
    def __str__(self):
        fields = {name: str(getattr(self, name)) for name in self._fields_}
        return f"{type(self).__name__}({self._vec}, {fields})"


BitField.Field = Field


class BitfieldArgs:
    def __init__(self, arg):
        assert isinstance(arg, int)
        self.offset: int = arg

    @_intrinsic
    def __str__(self) -> str:
        return str(self.offset)

    def __hash__(self) -> int:
        return hash(self.offset)

    def __eq__(self, other: BitfieldArgs) -> bool:
        return self.offset == other.offset


class _BitFieldInst(BitField):
    # _offset_: int
    # _fields_: dict[str, type]

    @_intrinsic
    def __class_getitem__(cls, arg):
        if isinstance(arg, slice):
            assert isinstance(arg.start, int), "slice start must be an integer"
            assert isinstance(arg.stop, int), "slice stop must be an integer"
            assert arg.step is None, "slice step may not be used"
            assert (
                arg.start - arg.stop + 1 == cls._width_
            ), f"slice width ({arg.start}:{arg.stop}=={arg.start - arg.stop + 1}) does not match BitField width {cls._width_}"
            arg = arg.stop

        assert isinstance(arg, int), "expected slice or integer"
        subclasses = cls._cohdlstd_subclasses

        if arg in subclasses:
            return subclasses[arg]

        newtype = type(
            f"{cls.__name__}[{arg}]",
            (subclasses[None],),
            {"_offset_": arg, "_cohdlstd_is_offset_bitfield": False},
        )
        subclasses[arg] = newtype
        return newtype

    @_intrinsic
    def __init_subclass__(cls):
        if "_cohdlstd_is_bitfield_base" in cls.__dict__:
            return

        if hasattr(cls, "_cohdlstd_bitfield_init_done"):
            assert (
                cls._cohdlstd_is_offset_bitfield == False
            ), "BitFields may not be derived"
            cls._cohdlstd_is_offset_bitfield = True
            return

        fields = {}
        for name, value in typing.get_type_hints(cls).items():
            assert issubclass(
                value, (FieldBit, FieldBitVector, _BitFieldInst)
            ), f"invalid BitField element type {value} of element '{name}'"
            fields[name] = value

        cls._fields_ = fields
        setattr(cls, "_cohdlstd_bitfield_init_done", True)
        setattr(cls, "_cohdlstd_subclasses", {None: cls})



from __future__ import annotations

from cohdl import Port, BitVector, Signed, Unsigned
from cohdl import Entity

from cohdl import std

# obtain generic type and use it to produce a generic record
T = std.template_arg.Type


class Coord(std.Record[T]):
    x: T
    y: T

    def __add__(self, other: Coord) -> Coord:
        return type(self)(x=self.x + other.x, y=self.y + other.y)


class MyEntity(Entity):
    x_in = Port.input(BitVector[32])
    y_in = Port.input(BitVector[32])

    x_out_signed = Port.output(BitVector[32])
    y_out_signed = Port.output(BitVector[32])

    x_out_unsigned = Port.output(BitVector[16])
    y_out_unsigned = Port.output(BitVector[16])

    def architecture(self):
        # specialize the generic
        out_signed = std.Ref[Coord[Signed[32]]](self.x_out_signed, self.y_out_signed)

        @std.concurrent
        def logic_signed():
            nonlocal out_signed

            # specialize generic Coord with Signed[32]
            coord_in = Coord[Signed[32]](self.x_in, self.y_in)

            # coord is not a signal
            # but can be assigned like one
            out_signed <<= coord_in

        @std.concurrent
        def logic_unsigned():

            # specialize generic Coord with Unsigned[16]
            a = Coord[Unsigned[16]](self.x_in[15:0], self.y_in[15:0])
            b = Coord[Unsigned[16]](self.x_in[31:16], self.y_in[31:16])

            # create a new instance of Coord from constant arguments
            c = Coord[Unsigned[16]](3, 4)

            sum = a + b + c

            self.x_out_unsigned <<= sum.x
            self.y_out_unsigned <<= sum.y


vhdl = std.VhdlCompiler.to_string(MyEntity)
print(vhdl)



from __future__ import annotations

from types import TracebackType

from .source_location import SourceLocation


class VirtualFrame:
    def __init__(
        self, location: SourceLocation, scope: dict, parent: VirtualFrame | None = None
    ):
        self._location = location
        self._scope = scope
        self._parent = parent

    def location(self):
        # define location as a function so it can be overwritten
        return self._location

    def ignore(self, prev: VirtualFrame) -> bool:
        return prev.location() == self.location()

    def _add_single(
        self,
        prev: VirtualFrame | None = None,
        prev_tb: TracebackType | None = None,
    ):
        # helper function to insert stack frames from traced python code
        # to the exception traceback

        if prev is not None:
            if self.ignore(prev):
                return prev_tb

        loc = self.location()

        function_name = loc.function if loc.function is not None else "UNKNOWN_FUNCTION"

        # It is not possible to construct the python builtin FrameType from
        # user code. As a workaround we create a fake by constructing a string of python code
        # that contains a function with the same name. The fake function raises an assertion
        # so we can extract the frame object after a call to `exec`.
        code_str = f"def {function_name}():\n\traise AssertionError('dummy')\n{function_name}()\n"

        compiled_fn = compile(code_str, filename=loc.file, mode="exec")

        try:
            # Use the fake function of create a frame object with the correct
            # function name, line number, filename and local/global variables
            exec(compiled_fn, self._scope)
        except BaseException as dummy_err:
            new_frame = dummy_err.__traceback__.tb_next.tb_next.tb_frame

            for name, val in self._scope.items():
                new_frame.f_locals[name] = val

            return TracebackType(prev_tb, new_frame, 1, loc.line)

    def apply_to_exception(self, exception: Exception, extend=False):
        """
        Replaces the existing traceback of the given `exception` with one
        constructed from this virtual frame.
        When `extend` is set to true the existing traceback is not discarded.
        Instead the virtual traceback is added to it.
        """

        current_frame = self
        prev_frame = None

        if extend:
            custom_tb = exception.__traceback__
        else:
            custom_tb = None

        while current_frame is not None:
            custom_tb = current_frame._add_single(prev_frame, custom_tb)
            prev_frame = current_frame
            current_frame = current_frame._parent

        exception.__traceback__ = custom_tb

    def relative(self, offset: int, parent=None):
        return VirtualFrame(
            self.location().relative(offset),
            self._scope,
            self if parent is None else parent,
        )



from __future__ import annotations

from cohdl import Bit, BitVector, Unsigned, Signal, Null, Full
from cohdl import std


class Register:
    def __init__(self, offset):
        assert offset % 4 == 0
        self.offset = offset

    async def on_read(self):
        return Null

    async def on_write(self, data, mask=Full):
        pass


class RegisterSet:
    def __init__(self, content: list[Register | RegisterSet], offset=0):
        self._content = content
        self._offset = offset

    def get_registers(self, offset=0):
        result = []

        for entry in self._content:
            if isinstance(entry, Register):
                entry.offset += self._offset + offset
                result.append(entry)
            elif isinstance(entry, RegisterSet):
                entry._offset += self._offset + offset
                result.extend(entry.get_registers())

        return result


class RegisterDevice:
    def __init__(
        self, interface: MemoryInterface, registers: list[Register | RegisterSet]
    ):
        flat_registers: list[Register] = []

        for entry in registers:
            if isinstance(entry, Register):
                flat_registers.append(entry)
            else:
                flat_registers.extend(entry.get_registers())

        for reg in flat_registers:
            assert reg.offset < 2 ** interface.addr_width()

        async def on_read(addr):
            aligned_addr = addr.unsigned

            result = Variable[BitVector[32]](Null)

            for reg in flat_registers:
                if reg.offset == aligned_addr:
                    result @= await std.as_awaitable(reg.on_read)
                    break

            return result

        async def on_write(addr, data, mask):
            aligned_addr = addr.unsigned

            for reg in flat_registers:
                if reg.offset == aligned_addr:
                    await std.as_awaitable(reg.on_write, data, mask)
                    break

        interface.run_request_handler(on_read, on_write)


class SimpleRegister(Register):
    def __init__(
        self,
        offset: int,
        sig: Signal[BitVector[32]],
        read_only: bool = False,
        write_only: bool = False,
    ):
        super().__init__(offset)

        self.sig = sig
        self.read_only = read_only
        self.write_only = write_only

    async def on_read(self):
        if self.write_only:
            return Null
        else:
            return self.sig

    async def on_write(self, data, mask=Full):
        if self.read_only:
            pass
        else:
            self.sig <<= std.apply_mask(self.sig, data, std.stretch(mask, 8))


class PartialRegister(Register):
    def __init__(
        self,
        offset: int,
        sig: Signal[BitVector],
        read_only: bool = False,
        write_only: bool = False,
    ):
        super().__init__(offset)

        self.sig = sig
        self.read_only = read_only
        self.write_only = write_only

    async def on_read(self):
        if self.write_only:
            return Null
        else:
            return self.sig.unsigned.resize(32)

    async def on_write(self, data: Signal[BitVector[32]], mask=Full):
        if self.read_only:
            pass
        else:
            self.sig <<= data.lsb(self.sig.width)


class WaitRegister(Register):
    async def on_write(self, data: Unsigned, mask=Full):
        await std.wait_for(data.unsigned)


a = "asdf" "sadf"


class reg:
    class R:
        ...

    class W:
        ...

    class RW:
        ...

    class Field:
        def __class_getitem__(cls, *args) -> type[reg.Field]:
            ...

        ...

    class FieldS(Field):
        ...

    class FieldU(Field):
        ...

    class Register:
        ...

    class FieldRegister:
        def _data_view_(self, input, mask=Full) -> SpiMaster.CtrlRegister:
            ...

        def _super_init_(self):
            self._raw_ = Signal[BitVector[32]]()

            @std.concurrent
            def logic():
                self._raw_[5:2] <<= self.count
                self._raw_[8] <<= self.active

        def on_read_default(self):
            return self._raw_

        def on_write_default(self, data, mask=Full):
            field_view = self.field_view(data, mask)

    class RegisterDevice:
        ...


class RgbLed(reg.Register):
    "This register device controls the rgb leds"

    r: reg.FieldU[7:0]
    g: reg.FieldU[15:8]
    b: reg.FieldU[23:16]

    def __init__(self, ctx: std.SequentialContext, rgb):
        cnt = Signal[Unsigned[8]](Null)

        @ctx
        async def proc():
            await std.wait_for(std.us(5))
            cnt.next = cnt + 1
            rgb.red <<= cnt < self.r
            rgb.green <<= cnt < self.g
            rgb.blue <<= cnt < self.b


class SpiMaster(reg.RegisterDevice):
    class CtrlRegister(reg.Register):
        "This register controls the spi communication"

        count: reg.Field[5:2, 0]["asdf"]
        active: reg.Field[8]["set to '1' while a spi transaction is in progress"]

        def __init__(self):
            self.transaction_active = std.SyncFlag()

            std.concurrent_eval(self.active, self.transaction_active.is_set)

        def on_write(self, data, mask=Full):
            view = self._data_view_(data, mask)

            if view.active:
                assert not self.transaction_active.is_set()
                self.transaction_active.set()

            self.count <<= view.count

    ctrl: CtrlRegister[0]
    send_low: reg.W[4]
    send_high: reg.W[8]
    receive_low: reg.R[12]
    receive_high: reg.R[16]

    def __init__(
        self,
        ctx: std.SequentialContext,
        spi: std.spi.Spi,
        clk_period=std.Duration,
    ):
        super().__init__()

        master = std.spi.SpiMaster(ctx, spi, clk_period=clk_period)

        send_data = Signal[BitVector[64]](Null)
        receive_data = Signal[BitVector[64]](Null)

        @std.concurrent
        def logic():
            self.receive_low <<= receive_data[31:0]
            self.receive_high <<= receive_data[63:32]
            send_data[31:0] <<= self.send_low
            send_data[63:32] <<= self.send_high

        @ctx
        async def proc():
            nonlocal receive_data

            async with self.ctrl.transaction_active:
                receive_data <<= await master.parallel_transaction(
                    send_data, self.ctrl.count
                )


class Wait(reg.Register):
    async def on_write(self, value: BitVector, mask=Full):
        std.wait_for(value.unsigned)


class Interface(reg.RegisterDevice):
    """
    Collection of all hardware interfaces.
    """

    rgb_a: RgbLed[0x100]
    rgb_b: RgbLed[0x104]

    spi: SpiMaster[0x200]

    def __init__(self, ctx, rgb_a, rgb_b, spi, spi_period):
        self.rgb_a = RgbLed(ctx, rgb_a)
        self.rgb_b = RgbLed(ctx, rgb_b)
        self.spi = SpiMaster(ctx, spi, spi_period)


class Args:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs


def _main_():
    con = interconnect.reserve(0x10000000, 10)

    con.registers()

    all_regs(
        {
            0: reg.R(switches, "switches"),
            4: reg.W(leds, "leds"),
            8: Wait(),
            12: reg.R(buttons, "buttons"),
            0x100: ethernet.EthernetDevice(
                ctx=eth_ctx,
                crsdv=eth.crsdv,
                rxd0=eth.rxd0,
                rxd1=eth.rxd1,
                txen=eth.txen,
                txd0=eth.txd0,
                txd1=eth.txd1,
                uart_tx=uart_tx,
            ),
            0x200: ethernet.UartDevice(
                ctx=eth_ctx,
                rx=uart_rx,
                tx=uart_tx,
                baud=std.uart.Baud.BAUD_115200,
            ),
        }
    )

    class AllRegisters(reg.RegisterDeviceXX):
        SW = reg.R[0](switches)
        LED = reg.W[4](leds)
        WAIT = Wait[8]()
        BUTTONS = reg.R[12](buttons_vec)
        ETH = ethernet.EthernetDevice[0x100](
            ctx=eth_ctx,
            crsdv=eth.crsdv,
            rxd0=eth.rxd0,
            rxd1=eth.rxd1,
            txen=eth.txen,
            txd0=eth.txd0,
            txd1=eth.txd1,
            uart_tx=uart_tx,
        )

        UART = ethernet.UartDevice[0x200](
            ctx=eth_ctx,
            rx=uart_rx,
            tx=uart_tx,
            baud=std.uart.Baud.BAUD_115200,
        )

        SEVENT_SEG = ethernet.HexDisplay[0x300](
            ctx=eth_ctx, anodes=seven_seg.anodes, cathodes=seven_seg.cathodes
        )

        RGB_A = RgbLed[0x400](ctx=eth_ctx, rgb=rgb_1)
        RGB_B = RgbLed[0x404](ctx=eth_ctx, rgb=rgb_2)

        SPI = SpiMaster[0x800](
            ctx=eth_ctx, spi=acc.spi, clk_period=std.MHz(2.5).period()
        )

    RegisterDevice(
        interconnect.reserve(0x10000000, 10),
        [
            PartialRegister(0x00, switches, read_only=True),
            PartialRegister(0x04, leds, write_only=True),
            WaitRegister(0x08),
            PartialRegister(0x0C, button_vec, read_only=True),
            ethernet.EthernetDevice(
                0x10,
                ctx=eth_ctx,
                crsdv=eth.crsdv,
                rxd0=eth.rxd0,
                rxd1=eth.rxd1,
                txen=eth.txen,
                txd0=eth.txd0,
                txd1=eth.txd1,
                uart_tx=uart_tx,
            ),
            ethernet.UartDevice(
                0x20,
                ctx=eth_ctx,
                rx=uart_rx,
                tx=uart_tx,
                baud=std.uart.Baud.BAUD_115200,
            ),
            ethernet.HexDisplay(
                0x100, ctx=eth_ctx, anodes=seven_seg.anodes, cathodes=seven_seg.cathodes
            ),
            ethernet.RgbLed(0x180, ctx=eth_ctx, rgb=rgb_1),
            ethernet.RgbLed(0x184, ctx=eth_ctx, rgb=rgb_2),
            ethernet.SpiMaster(
                0x200, ctx=eth_ctx, spi=acc.spi, clk_period=std.MHz(2.5).period()
            ),
        ],
    )


if False:
    ctrl: CtrlRegister[0]

    send_low: reg.SimpleRegister.Wr[
        4,
        "dasd asist sfiasrlasfljsdalfölasd asdflsladjoislfldsafldn welsd wejjoinvhhsdf fsds",
    ]

    send_high: reg.SimpleRegister.Wr[8]

    receive_low: reg.SimpleRegister.Rd[12]
    receive_high: reg.SimpleRegister.Rd[16]

    def __init__(
        self,
        ctx: std.SequentialContext,
        spi: std.spi.Spi,
        clk_period=std.Duration,
    ):
        master = std.spi.SpiMaster(ctx, spi, clk_period=clk_period)

        send_data = Signal[BitVector[64]](Null)
        receive_data = Signal[BitVector[64]](Null)
        ctrl_register = Signal[BitVector[32]](Null)

        transaction_active = std.SyncFlag()

        @ctx
        async def proc():
            nonlocal receive_data

            async with transaction_active:
                receive_data <<= await master.parallel_transaction(
                    send_data, ctrl_register[5:0].unsigned
                )

        SimpleRegister()
        super().__init__(
            [
                CtrlRegister(0x00),
                SimpleRegister(0x04, send_data[31:0], write_only=True),
                SimpleRegister(0x08, send_data[63:32], write_only=True),
                SimpleRegister(0x0C, receive_data[31:0], read_only=True),
                SimpleRegister(0x10, receive_data[63:32], read_only=True),
            ],
        )



from __future__ import annotations


from ._primitive_type import is_primitive_type, _PrimitiveType
from ._intrinsic import _intrinsic
from ._enum import Enum
from ._boolean import Null, Full


class _MetaArray(type):
    _SubTypes: dict[tuple, type]

    _elemtype_: type
    _count_: int

    @_intrinsic
    def __len__(cls):
        return cls._count_

    @_intrinsic
    def __getitem__(cls, slice):
        assert (
            isinstance(slice, tuple) and len(slice) == 2
        ), "cohdl.Array[] requires two arguments [DATA_TYPE, SIZE]"

        elemtype, count = slice

        assert isinstance(elemtype, type)
        assert isinstance(count, int)

        assert is_primitive_type(elemtype)

        content_type = (elemtype, count)

        if content_type in cls._SubTypes:
            return cls._SubTypes[content_type]

        new_type = type(
            cls.__name__, (cls,), {"_elemtype_": elemtype, "_count_": count}
        )
        cls._SubTypes[content_type] = new_type
        return new_type

    @_intrinsic
    def __str__(cls):
        return f"{cls.__name__}[{cls._elemtype_}, {cls._count_}]"


class Array(_PrimitiveType, metaclass=_MetaArray):
    _SubTypes = {}
    _elemtype_: type
    _count_: int

    @_intrinsic
    def __init__(self, val=None):
        elemtype = self._elemtype_

        if isinstance(val, Array):
            val = val._value

        if isinstance(val, (list, tuple)):
            assert (
                len(val) <= self._count_
            ), "more default arguments than array elements"
            self._value = [elemtype(v) for v in val]
        elif val is Null or val is Full:
            self._value = [elemtype(val) for _ in range(self._count_)]
        else:
            assert val is None, "invalid array type"
            self._value = None

    @_intrinsic
    def __len__(self):
        return self._count_

    @_intrinsic
    def __getitem__(self, index: int):
        elem_type = self._elemtype_

        if issubclass(elem_type, Enum):
            first, *rest = elem_type._member_map_.values()
            return elem_type(first)

        return elem_type()

    @_intrinsic
    def __setitem__(self, slice, arg):
        raise AssertionError(
            "array set item not supported outside synthesizable contexts"
        )

    @_intrinsic
    def _assign(self, value):
        assert isinstance(value, Array)
        assert (
            value._count_ == self._count_
        ), f"width of source array does not match width of target ({value._count_} != {self._count_})"

        self_val = self._elemtype_()

        for val in value._value:
            # Dummy assignment, does not actually assign to array members
            # to avoid creating potentially large number of elements.
            # Only used to ensure, that the assignment is sound.
            self_val._assign(val)

    @_intrinsic
    def copy(self) -> Array:
        return type(self)(self._value)

    @_intrinsic
    def __hash__(self):
        return hash(self._elemtype_, self._count_)

    @_intrinsic
    def __str__(self) -> str:
        return self.__repr__()

    @_intrinsic
    def __repr__(self) -> str:
        return f"{type(self)}()"



from . import _core as core

from ._core import (
    Bit,
    BitState,
    BitOrder,
    BitVector,
    Unsigned,
    Integer,
    Boolean,
    Signed,
    Variable,
    TypeQualifierBase,
    TypeQualifier,
    Signal,
    Temporary,
    Attribute,
    Block,
    Entity,
    Port,
    Generic,
    true,
    false,
    Null,
    Full,
    _NullFullType,
)

from ._core import _ir as ir
from ._core import enum

from ._core import reset_context, reset_pushed, sensitivity
from ._core import (
    comment,
    select_with,
    always,
    expr,
    expr_fn,
    evaluated,
    static_assert,
    coroutine_step,
)
from ._core import rising_edge, falling_edge, high_level, low_level
from ._core import Array
from ._core._intrinsic import _intrinsic as constexpr
from ._core._intrinsic import _intrinsic as consteval
from ._core._intrinsic import _intrinsic as pyeval
from ._core import AssignMode
from ._core import _BitSignalEvent as BitSignalEvent
from ._core import sequential_context, concurrent_context
from ._core import is_primitive, is_primitive_type

from ._core import _InlineCode, inline_raw, vhdl

from ._compiler.frontend import use_pretty_traceback

from . import std
from . import utility



from __future__ import annotations

import inspect
import sys

import enum
from dataclasses import dataclass

from cohdl._core._intrinsic import _intrinsic
from ._core_utility import nop


class _TemplateMode(enum.Enum):
    ROOT = enum.auto()
    DECLARATION = enum.auto()
    SPECIALIZED = enum.auto()


class _TemplateMeta:
    def __init__(
        self,
        mode: _TemplateMode,
        argtype: type | None = None,
        instances: dict | None = None,
        original_class_getitem=None,
        annotations: dict | None = None,
        arg=None,
    ):
        self.arg = arg
        self.mode = mode
        self.argtype = argtype
        self.instances: dict[type, dict[any, type]] = instances
        self.original_class_getitem = original_class_getitem
        self.annotations = annotations

    def instance_exists(self, specialized_type: type, template_arg):
        specialized_cache = self.instances.setdefault(specialized_type, {})
        return template_arg in specialized_cache

    def get_instance(self, specialized_type: type, template_arg):
        return self.instances[specialized_type][template_arg]

    def add_instance(self, specialized_type: type, template_arg, instance):
        specialized_cache = self.instances.setdefault(specialized_type, {})
        assert (
            template_arg not in specialized_cache
        ), "internal error: template arg already in cache"
        specialized_cache[template_arg] = instance


@_intrinsic
def class_getitem_specialize(cls: type[Template], args):
    meta = cls._template_meta_

    template_arg = meta.argtype(args)

    if meta.instance_exists(cls, template_arg):
        return meta.get_instance(cls, template_arg)

    # instantiate and add new type early
    # so nested template types are possible
    #
    # shadow __init_subclass__ with a std.nop
    # because the class is not completely initialized at this
    # point. Explicitly call the original __init_subclass__
    # at the end of this function.
    newtype = type(
        f"{cls.__name__}[{template_arg}]",
        (cls,),
        {"__init_subclass__": nop},
    )

    meta.add_instance(cls, template_arg, newtype)

    class_members = {}

    if hasattr(cls, "__class_getitem__"):
        class_members["__class_getitem__"] = cls.__class_getitem__

    template_annotations = {}

    # adopted from typing.get_type_hints
    for base in reversed(cls.__mro__):
        base_globals = getattr(sys.modules.get(base.__module__, None), "__dict__", {})
        ann = base.__dict__.get("__annotations__", {})

        base_locals = {
            name: val if val is not meta.argtype else template_arg
            for name, val in vars(base).items()
        }

        base_globals = {
            name: val if val is not meta.argtype else template_arg
            for name, val in base_globals.items()
        }

        base_globals, base_locals = base_locals, base_globals

        for name, value in ann.items():
            try:
                template_annotations[name] = eval(value, base_globals, base_locals)
            except TypeError as err:
                err.add_note(
                    f"maybe you are missing 'from __future__ import annotations' in {base.__module__}"
                )
                raise err
            except BaseException as err:
                raise err

    newdict = {
        **class_members,
        **template_annotations,
        "_template_meta_": _TemplateMeta(
            mode=_TemplateMode.SPECIALIZED,
            argtype=meta.argtype,
            instances=None,
            original_class_getitem=None,
            annotations=template_annotations,
            arg=template_arg,
        ),
    }

    for name, value in newdict.items():
        setattr(newtype, name, value)

    newtype._template_specialize_()

    # remove __init_subclass__ defined at the start
    # of this function
    delattr(newtype, "__init_subclass__")
    assert hasattr(newtype, "__init_subclass__")
    newtype.__init_subclass__()

    return newtype


class Template:
    _template_meta_ = _TemplateMeta(_TemplateMode.ROOT)

    @_intrinsic
    def __class_getitem__(cls, args):
        assert (
            cls._template_meta_.mode is _TemplateMode.ROOT
        ), "internal error: expected template mode ROOT"
        argtype = args
        assert isinstance(args, type), "Template expects a type argument"
        assert (
            argtype.__hash__ is not object.__hash__
        ), f"TemplatArg({argtype}) must define a hash function"
        assert (
            argtype.__eq__ is not object.__eq__
        ), f"TemplatArg({argtype}) must define an equality comparison operator"

        return type(
            "_TemplateDeclaration",
            (cls,),
            {
                "__class_getitem__": class_getitem_specialize,
                "_template_meta_": _TemplateMeta(
                    mode=_TemplateMode.DECLARATION,
                    argtype=argtype,
                    instances={},
                    original_class_getitem=None,
                ),
            },
        )

    @_intrinsic
    def __new__(cls, *args, **kwargs):
        meta = cls._template_meta_

        assert (
            meta.mode is not _TemplateMode.ROOT
        ), "unconstrained std.Template may not be instantiated"

        if meta.mode is _TemplateMode.SPECIALIZED:
            return super().__new__(cls)

        result_type = cls._template_deduce_(*args, **kwargs)
        result = object.__new__(result_type)
        return result

    @classmethod
    def _template_specialize_(cls):
        pass


@_intrinsic
def _unpack_init(self, real_init, arg):
    if isinstance(arg, tuple):
        real_init(self, *arg)
    else:
        real_init(self, arg)


class _TemplateArg:
    @property
    @_intrinsic
    def Type(self):
        class _Type:
            def __new__(cls, arg):
                assert isinstance(arg, type)
                return arg

            def __hash__(self):
                raise AssertionError(
                    "should never be called since __new__ never prevents this class from being instantiated"
                )

            def __eq__(self, other):
                raise AssertionError(
                    "should never be called since __new__ never prevents this class from being instantiated"
                )

        return _Type

    @_intrinsic
    def __call__(self, cls):
        dataclassed = dataclass(cls, unsafe_hash=True)

        real_init = dataclassed.__init__

        def __init__(self, arg):
            _unpack_init(self, real_init, arg)

        dataclassed.__init__ = __init__

        return dataclassed


template_arg = _TemplateArg()



from __future__ import annotations

from cohdl._core._intrinsic import _intrinsic
from cohdl._core import TypeQualifierBase, is_primitive_type


class _Prefix:
    _existing_prefix: dict[str, int] = {}
    _prefix_scope: list[_Prefix] = []

    @staticmethod
    def _parent_prefix():
        if len(_Prefix._prefix_scope) == 0:
            return None
        return _Prefix._prefix_scope[-1]

    @_intrinsic
    def __init__(self, prefix: str, subprefix=False):
        if not subprefix and len(_Prefix._prefix_scope) != 0:
            prefix = _Prefix._prefix_scope[-1].name(prefix)

        lower = prefix.lower()
        cnt = _Prefix._existing_prefix.get(lower, 0)
        _Prefix._existing_prefix[lower] = cnt + 1

        if cnt == 0:
            self._prefix = prefix
        else:
            if prefix.endswith("_"):
                self._prefix = f"{prefix}{cnt}"
            else:
                self._prefix = f"{prefix}_{cnt}"

        self._scope_active = 0
        self._used_names = []

    @_intrinsic
    def subprefix(self, subname: str):
        return _Prefix(self.name(subname), subprefix=True)

    @_intrinsic
    def name(self, name: str | None) -> str:
        if name is None:
            return self._prefix

        lower = name.lower()
        assert (
            not lower in self._used_names
        ), f"name '{name}' already used with this prefix '{self._prefix}'"
        self._used_names.append(lower)
        return f"{self._prefix}_{name}"

    @_intrinsic
    def prefix_str(self) -> str:
        return self._prefix

    @_intrinsic
    def __enter__(self):
        if self._scope_active == 0:
            _Prefix._prefix_scope.append(self)
        self._scope_active += 1
        return self

    @_intrinsic
    def __exit__(self, type, value, traceback):
        assert self._scope_active, "attempt to exit inactive scope"
        self._scope_active -= 1

        if self._scope_active == 0:
            assert _Prefix._prefix_scope.pop() is self
        return None


@_intrinsic
def prefix(base_name: str) -> _Prefix:
    return _Prefix(base_name)


@_intrinsic
def name(name: str) -> str:
    parent = _Prefix._parent_prefix()

    if parent is None:
        return name
    return parent.name(name)


class _NamedQualifier:
    @_intrinsic
    def __init__(self, type, qualifier, name):
        self.type = type

        if isinstance(qualifier, _NamedQualifier):
            self.qualifier = qualifier.qualifier

            if name is None:
                self.prefix = qualifier.prefix
            else:
                self.prefix = qualifier.prefix.subprefix(name)
        else:
            self.qualifier = qualifier
            assert name is not None
            self.prefix = prefix(name)

    def __call__(self, *args, **kwargs):
        if self.type is None:
            assert len(args) == 1 and len(kwargs) == 0
            T = type(TypeQualifierBase.decay(args[0]))
        else:
            T = self.type

        if is_primitive_type(T) or T is bool or T is int:
            name = self.prefix.name(kwargs.get("name", None))
            return self.qualifier[T](*args, **{**kwargs, "name": name})
        else:
            with self.prefix:
                return T(
                    *args,
                    **kwargs,
                    _qualifier_=_NamedQualifier(None, self, None),
                )

    @_intrinsic
    def __getitem__(self, T):
        return _NamedQualifier(T, self, None)


class _NamedQualifierBuilder:
    @_intrinsic
    def __getitem__(self, qualifier_name):
        qualifier, name = qualifier_name
        return _NamedQualifier(None, qualifier, name)


NamedQualifier = _NamedQualifierBuilder()



from __future__ import annotations
from abc import abstractmethod

from typing import Tuple

import cohdl
import typing
import ast

from cohdl._core import _type_qualifier

from cohdl._core import _intrinsic_operations as intr_op
from cohdl._core._intrinsic_operations import AssignMode

from cohdl._core._collect_ast_and_scope import FunctionDefinition
from cohdl._core._context import EntityInfo

from cohdl import Temporary, Signal

from cohdl import Bit
from cohdl.utility.code_writer import IndentBlock, TextBlock
from cohdl.utility.virtual_traceback import VirtualFrame

from ._value_branch import _ValueBranchHook, _ValueBranch, _MergedBranch


class AstVirtualFrame(VirtualFrame):
    """
    A virtual frame derived from an ast entry relative to the function frame
    it is used in
    """

    def __init__(
        self, ast: ast.AST, fn_frame: VirtualFrame, parent_frame: VirtualFrame | None
    ):
        super().__init__(fn_frame._location, fn_frame._scope, parent_frame)
        self._ast = ast

    def location(self):
        try:
            return self._location.relative(self._ast.lineno)
        except AttributeError:
            # self._ast might not have a lineno attribute
            # fall back to function location
            return self._location

    def ignore(self, prev: VirtualFrame) -> bool:
        # ignore statements in traceback because only function calls
        # and the expression, that caused the error are relevant
        if isinstance(
            self._ast,
            (
                list,
                ast.If,
                ast.While,
                ast.For,
                ast.With,
                ast.AsyncWith,
                ast.FunctionDef,
                ast.AsyncFunctionDef,
            ),
        ):
            return True

        # only generate one traceback frame per source line
        if self.location() == prev.location():
            return True

        return False


class Statement:
    def __init__(
        self,
        returns_always=False,
        return_paths: list[Return] | None = None,
        bound_statements: list[Statement] | None = None,
        contains_break=False,
        contains_continue=False,
    ):
        self._returns_always = returns_always
        self._return_paths = return_paths or []
        self._bound_statements = [] if bound_statements is None else bound_statements
        self._contains_break = contains_break
        self._contains_continue = contains_continue
        self._frame: AstVirtualFrame

    def add_bound_statement(self, stmt: Statement):
        self._bound_statements.append(stmt)

    def bound_statements(self) -> list[Statement]:
        return self._bound_statements

    def return_paths(self) -> list[Return]:
        return self._return_paths

    def returns(self) -> bool:
        return len(self._return_paths) != 0

    def returns_always(self) -> bool:
        return self._returns_always

    def contains_break(self) -> bool:
        return self._contains_break

    def contains_continue(self) -> bool:
        return self._contains_continue

    def dump_bound(self):
        return TextBlock(
            title="bound", content=[stmt.dump() for stmt in self._bound_statements]
        )

    @abstractmethod
    def dump(self) -> IndentBlock:
        raise AssertionError("abstract method called")


class Expression(Statement):
    def __init__(
        self,
        result,
        bound_statements: list[Statement] | None = None,
    ):
        assert not isinstance(result, _ValueBranch)
        self._result = result
        super().__init__(bound_statements=bound_statements)

    def result(self):
        return self._result

    @abstractmethod
    def dump(self) -> IndentBlock:
        raise AssertionError("abstract method called")


class Nop(Statement):
    def __init__(self):
        super().__init__(False, None)

    def dump(self) -> IndentBlock:
        return IndentBlock(title="Nop", content=self.dump_bound())


class Comment(Statement):
    def __init__(self, lines: list[str]):
        super().__init__(False, None)
        self.lines = lines

    def dump(self) -> IndentBlock:
        return IndentBlock(title="Comment", content=[*self.lines])


class StarredValue(Expression):
    def __init__(self, result, bound_statements: list[Statement] | None = None):
        super().__init__(result, bound_statements)


class Assign(Expression):
    def __init__(
        self,
        target,
        value,
        mode: AssignMode,
        bound_statements: list,
    ):
        self._target = target
        self._value = value
        self._mode = mode

        super().__init__(target, bound_statements)

    def target(self):
        return self._target

    def value(self):
        return self._value

    def dump(self) -> IndentBlock:
        return TextBlock(
            title=f"Assign (target={self._target}, value={self._value}, mode={self._mode})",
            content=self.dump_bound(),
        )


class InlineCode(Expression):
    def __init__(self, options: list[cohdl._InlineCode]):
        self.options = options
        self.expr_type = options[0].expr_type

        for option in options:
            assert (
                option.expr_type is self.expr_type
            ), "all languages must return the same type"

        if self.expr_type is None:
            super().__init__(self)
        else:
            super().__init__(Temporary[self.expr_type]())

    def dump(self):
        return IndentBlock(title="InlineCode", content=self.dump_bound())


class CodeBlock(Statement):
    def __init__(self, stmts: list[Statement]):
        self._stmts: list[Statement] = []

        returns_always = False
        return_paths: list[Return] = []

        contains_break = False
        contains_continue = False

        for stmt in stmts:
            if isinstance(stmt, CodeBlock):
                self._stmts.extend(stmt.statements())
            else:
                self._stmts.append(stmt)

            if stmt.contains_break():
                contains_break = True

            if stmt.contains_continue():
                contains_continue = True

            if stmt.returns():
                return_paths.extend(stmt.return_paths())

            if stmt.returns_always():
                returns_always = True
                break

        super().__init__(
            returns_always=returns_always,
            return_paths=return_paths,
            contains_break=contains_break,
            contains_continue=contains_continue,
        )

    def empty(self) -> bool:
        return len(self._stmts) == 0

    def statements(self) -> list[Statement]:
        return self._stmts

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="CodeBlock", content=[stmt.dump() for stmt in self._stmts]
        )


class Value(Expression):
    def __init__(self, value: typing.Any, bound_statements: list[Statement]):
        super().__init__(value, bound_statements=bound_statements)

    def dump(self):
        return IndentBlock(
            title=f"Value (value={self.result()})", content=self.dump_bound()
        )


class CohdlExpr(Value):
    """
    special case for cohdl.expr(...)
    """

    def dump(self):
        return IndentBlock(title=f"CohdlExpr (value={self.result()})", content=[])


class SignalAlias(Statement):
    """
    Signals, that are constructed in sequential contexts
    are not valid until the next cycle. To allow using the value
    immediately after its construction a Temporary is used in place of
    all ready during this initial state.
    """

    def __init__(self, signal, replacement, bound_statements: list[Statement]):
        super().__init__(bound_statements=bound_statements)
        self._signal = signal
        self._replacement = replacement

    def dump(self):
        return "SignalAlias"


class Boolean(Expression):
    """
    cast input value to a boolean (single bit)
    """

    def __init__(self, value: Expression):
        self._value = value.result()

        super().__init__(
            Temporary[bool](),
            bound_statements=[value],
        )

    def value(self):
        return self._value

    def dump(self):
        return IndentBlock(title="Boolean", content=[])


class If(Statement):
    def __init__(self, test: Expression, body: CodeBlock, orelse: CodeBlock):
        self._test = test
        self._body = body
        self._orelse = orelse

        returns_always = self._body.returns_always() and self._orelse.returns_always()
        return_paths = self._body.return_paths() + self._orelse.return_paths()

        super().__init__(
            returns_always,
            return_paths,
            contains_break=body.contains_break() or orelse.contains_break(),
            contains_continue=body.contains_continue() or orelse.contains_continue(),
        )

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"If",
            content=[
                IndentBlock(title="test", content=self._test.dump()),
                IndentBlock(title="body", content=self._body.dump()),
                IndentBlock(title="orelse", content=self._orelse.dump()),
            ],
        )


class BinOp(Expression):
    Operator = intr_op.BinaryOperator

    def __init__(
        self, op: BinOp.Operator, lhs: Expression, rhs: Expression, result
    ) -> None:
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

        super().__init__(result)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"BinOp (operator={self._op})",
            content=[
                IndentBlock(title="lhs", content=self._lhs.dump()),
                IndentBlock(title="rhs", content=self._rhs.dump()),
            ],
        )


class UnaryOp(Expression):
    Operator = intr_op.UnaryOperator

    def __init__(self, op: UnaryOp.Operator, arg: Expression, result):
        self._op = op
        self._arg = arg

        super().__init__(result, [])

    def arg(self):
        return self._arg.result()

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"UnaryOp (operator={self._op})",
            content=[IndentBlock(title="arg", content=self._arg.dump())],
        )


class Compare(Expression):
    Operator = intr_op.ComparisonOperator

    def __init__(self, op: Compare.Operator, lhs: Expression, rhs: Expression, result):
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

        super().__init__(result)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"SingleCompare (operator={self._op})",
            content=[
                IndentBlock(title="lhs", content=self._lhs.dump()),
                IndentBlock(title="rhs", content=self._rhs.dump()),
            ],
        )


class All(Expression):
    def __init__(self, conditions: list, bound_expressions: list[Expression]):
        if len(conditions) == 0:
            result = True
        else:
            result = _type_qualifier.Temporary[bool]()

        super().__init__(
            result,
            bound_statements=bound_expressions,
        )

        self._conditions = conditions

    def dump(self) -> str:
        return "all"


class Any(Expression):
    def __init__(self, conditions: list, bound_expressions: list[Expression]):
        if len(conditions) == 0:
            result = False
        else:
            result = _type_qualifier.Temporary[bool]()

        super().__init__(
            result,
            bound_statements=bound_expressions,
        )

        self._conditions = conditions

    def dump(self) -> str:
        return "any"


count = 0


class IfExpr(Expression):
    def __init__(self, test: Expression, body: Expression, orelse: Expression):
        global count
        self._test = test
        self._body = body
        self._orelse = orelse

        self._hook_orelse = _ValueBranchHook(name="hookorelse")
        count = count + 1
        self._hook_body = _ValueBranchHook(name="hookbody")
        count = count + 1

        result = _MergedBranch(
            [
                _ValueBranch(self._hook_body, body.result()),
                _ValueBranch(self._hook_orelse, orelse.result()),
            ]
        )

        super().__init__(result)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="IfExpr",
            content=[
                IndentBlock(title="test", content=self._test.dump()),
                IndentBlock(title="body", content=self._body.dump()),
                IndentBlock(title="orelse", content=self._orelse.dump()),
            ],
        )


class Return(Statement):
    def __init__(self, expr: Expression):
        self._hook = _ValueBranchHook(name="return")
        self._branch = _ValueBranch(self._hook, expr.result())

        # used by context managers
        # separate from normal _bound_statements
        # because __exit__ code must run after return value redirect
        self._final_bound_statements = []

        super().__init__(
            returns_always=True, return_paths=[self], bound_statements=[expr]
        )

    def bound_statements(self) -> list[Statement]:
        return super().bound_statements() + [
            Assign(redirect.target, redirect.source, AssignMode.AUTO, [])
            for redirect in self._hook.redirects
        ]

    def branch(self) -> _ValueBranch:
        return self._branch

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="Return", content=f"nr_redirects={len(self._hook.redirects)}"
        )


class While(Statement):
    def __init__(self, test: Expression, body: CodeBlock):
        self._test = test
        self._body = body

        super().__init__(
            False,
            [*body.return_paths()],
            bound_statements=[test],
            contains_break=False,  # while loop consumes contained break statements
            contains_continue=False,  # while loop consumes contained continue statements
        )

    def uses_break(self) -> bool:
        return self._body.contains_break()

    def uses_continue(self) -> bool:
        return self._body.contains_continue()

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="While",
            content=[
                IndentBlock(title="test", content=self._test.dump()),
                IndentBlock(title="body", content=self._body.dump()),
                IndentBlock(title="orelse", content=self._orelse.dump()),
            ],
        )


class Break(Statement):
    def __init__(self):
        super().__init__(contains_break=True)

    def dump(self) -> IndentBlock:
        return IndentBlock(title="break")


class Continue(Statement):
    def __init__(self):
        super().__init__(contains_continue=True)

    def dump(self) -> IndentBlock:
        return IndentBlock(title="continue")


class Await(Expression):
    def __init__(
        self, expr: Expression, primitive: bool, expr_before: list | None = None
    ):
        expr_result = expr.result()
        self._awaitable_primitive = primitive
        self._expr_before = expr_before if expr_before is not None else []
        super().__init__(expr_result, bound_statements=[expr])

    def dump(self) -> IndentBlock:
        return IndentBlock(title="Await", content=self.dump_bound())


class Call(Expression):
    def __init__(self, code: CodeBlock):
        self._code = code
        result = _MergedBranch([stmt.branch() for stmt in code.return_paths()])

        if code.returns() and result is not None:
            assert code.returns_always()

        super().__init__(result)

    def code(self) -> CodeBlock:
        return self._code

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"Call -> {self.result()}", content=self._code.dump())


class ResetInstance(Statement):
    def __init__(self, obj):
        assert isinstance(obj, _type_qualifier.TypeQualifier) and obj.has_default()
        self._obj = obj
        super().__init__()

    def get_obj(self):
        return self._obj

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"ResetInstance {self._obj}")


class FunctionDef(Statement):
    def __init__(self, fn_def: FunctionDefinition):
        bound_statements = []
        super().__init__(bound_statements=bound_statements)

        self._fn_def = fn_def

    def bind_args(self, args, kwargs):
        return self._fn_def.bind_args(args, kwargs)

    def dump(self) -> IndentBlock:
        return IndentBlock(title=f"FunctionDef ({self._fn_def.name()})")


#
#
# special intrinsic
#
#


class SelectWith(Expression):
    def __init__(
        self,
        arg,
        branches: list[Tuple[Value, typing.Any]],
        default: typing.Any | None,
    ):
        self._arg = arg
        self._conditions = [branch[0] for branch in branches]
        self._branches = branches
        self._default = default

        self._branch_hooks: list[_ValueBranchHook] = []

        branch_values = []

        for branch in branches:
            try:
                arg == branch[0]
            except:
                pass
                # raise AssertionError(f"argument not comparable to all branches")

            hook = _ValueBranchHook()
            self._branch_hooks.append(hook)
            branch_values.append(_ValueBranch(hook, branch[1]))

        if default is None:
            self._default_hook = None
        else:
            self._default_hook = _ValueBranchHook()
            branch_values.append(_ValueBranch(self._default_hook, default))

        super().__init__(
            _MergedBranch(branch_values),
            bound_statements=[],
        )

    def dump(self) -> IndentBlock:
        if self._default is None:
            default = []
        else:
            default = [IndentBlock(title="[default]", content=f"{self._default}")]

        return IndentBlock(
            title="SelectWith",
            content=[
                IndentBlock(title=f"[{cond}]", content=f"{value}")
                for cond, value in self._branches
            ]
            + default,
        )


class CondSelect(Statement):
    def __init__(
        self, branches: list[Tuple[Expression, CodeBlock]], default: CodeBlock | None
    ):
        self._branches = branches
        self._default = default

        code_default = [] if default is None else [default]
        code_branches = [branch[1] for branch in branches]

        returns = any(branch.returns() for branch in code_branches + code_default)

        return_paths = []

        if returns:
            # if any branch returns all branches must always return
            assert all(
                branch.returns_always() for branch in code_branches + code_default
            )

            for branch in code_branches + code_default:
                return_paths.extend(branch.return_paths())

        super().__init__(returns, return_paths)

    def dump(self) -> IndentBlock:
        if self._default is not None:
            return IndentBlock(
                title="CondSelect",
                content=[
                    IndentBlock([expr.dump(), code.dump()])
                    for expr, code in self._branches
                ]
                + [IndentBlock(title="[default]", content=self._default.dump())],
            )
        else:
            return IndentBlock(
                title="CondSelect",
                content=[
                    IndentBlock([expr.dump(), code.dump()])
                    for expr, code in self._branches
                ],
            )


class Statemachine(Statement):
    def __init__(self, body: CodeBlock, name: str):
        super().__init__()
        self._body = body
        self._name = name

    def dump(self) -> IndentBlock:
        return IndentBlock(title="Statemachine", content=self._body.dump())


class ResetContext(Statement):
    def dump(self):
        return IndentBlock(title="ResetContext", content=[])


class ResetPushed(Statement):
    def dump(self):
        return IndentBlock(title="ResetPushed", content=[])


#
# Assertation
#


class Assert(Statement):
    def __init__(self, cond: Expression, message: str | None):
        super().__init__(bound_statements=cond.bound_statements())

        self._cond = cond
        self._msg = message

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="Assertation",
            content=[
                IndentBlock(title="condition", content=self._cond.dump()),
                IndentBlock(title="message", content=self._msg),
            ],
        )


#
#
#
# instances
#
#
#


class Block:
    def __init__(self, info: EntityInfo, subblocks, contexts, attributes):
        self._info = info
        self._subblocks = subblocks
        self._contexts = contexts
        self._attributes = attributes

    def subblocks(self):
        return self._subblocks

    def contexts(self):
        return self._contexts

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"Block",
            content=[sub.dump() for sub in self._subblocks],
        )


class EntityTemplate(Block):
    def __init__(self, info: EntityInfo, subblocks: list | None, contexts: list | None):
        if info.extern:
            assert (
                subblocks is None and contexts is None
            ), "extern blocks cannot contain subbocks or synthesizable contexts"

            super().__init__(info, [], [], {})
        else:
            super().__init__(info, subblocks, contexts, info.attributes)

    def extern(self):
        return self._info.extern

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"EntityTemplate ({self._info.name})",
            content=[
                IndentBlock(
                    title="Ports",
                    content=[repr(port) for port in self._info.ports],
                ),
                IndentBlock(
                    title="Subinstances",
                    content=[inst.dump() for inst in self.subinstances()],
                ),
            ],
        )


class Entity(Block):
    def __init__(
        self,
        template: EntityTemplate,
        port_definitions: dict[str, Signal],
        generic_definitions: dict[str, Any],
    ):
        for name, port in template._info.ports.items():
            port._assign_(port_definitions[name], AssignMode.NEXT)

        super().__init__(template._info, [], [], template._info.attributes)
        self._template = template
        self._port_definitions = port_definitions
        self._generic_definitions = generic_definitions

    def template(self) -> EntityTemplate:
        return self._template

    def port_definitions(self) -> dict[str, Signal]:
        return self._port_definitions

    def generic_definitions(self):
        return self._generic_definitions

    def instantiate(self) -> Block:
        raise NotImplementedError()

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title="EntityWrapper",
            content=[
                IndentBlock(title="template", content=self._template.dump()),
                IndentBlock(
                    title="port definitions",
                    content=[
                        f"{name}: {value}"
                        for name, value in self._port_definitions.items()
                    ],
                ),
            ],
        )


#
# synthesizable contexts
#


class Context:
    def __init__(self, name: str, code: CodeBlock, attributes: dict, source_location):
        self._name = name
        self._code = code
        self._attributes = attributes
        self._source_loc = source_location

    def name(self):
        return self._name

    def attributes(self):
        return self._attributes

    def code(self):
        return self._code

    def source_location(self):
        return self._source_loc


class Concurrent(Context):
    def __init__(self, name, code: CodeBlock, attributes: dict, source_location):
        super().__init__(name, code, attributes, source_location)

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"Concurrent (name={self._name})",
            content=self._code.dump(),
        )


class Sequential(Context):
    def __init__(
        self,
        name,
        code: CodeBlock,
        always_expr: list,
        sensitivity: list,
        attributes: dict,
        source_location,
    ):
        super().__init__(name, code, attributes, source_location)
        self._always_expr = always_expr
        self._sensitivity = sensitivity

    def sensitivity(self):
        return self._sensitivity

    def always_expr(self):
        return self._always_expr

    def dump(self) -> IndentBlock:
        return IndentBlock(
            title=f"Sequential (name={self._name})",
            content=self._code.dump(),
        )



from __future__ import annotations

from ._bit_vector import BitVector
from ._integer import Integer
from ._unsigned import Unsigned

class Signed(BitVector):
    def __init__(
        self,
        val: None | BitVector | str | int = None,
    ): ...
    @classmethod
    def min(self) -> int: ...
    @classmethod
    def max(self) -> int: ...
    def to_int(self) -> int: ...
    def add(self, rhs: Signed | int | Integer, target_width=None) -> Signed: ...
    def sub(self, rhs: Signed | int | Integer, target_width=None) -> Signed: ...
    def __add__(self, rhs: Signed | int | Integer) -> Signed: ...
    def __radd__(self, lhs: int | Integer) -> Signed: ...
    def __sub__(self, rhs: Signed | int | Integer) -> Signed: ...
    def __rsub__(self, lhs: int | Integer) -> Signed: ...
    def __mul__(self, rhs: Signed | int | Integer) -> Signed: ...
    def __neg__(self) -> Signed: ...
    def __and__(self, other: Signed) -> Signed: ...
    def __or__(self, other: Signed) -> Signed: ...
    def __xor__(self, other: Signed) -> Signed: ...
    def __eq__(self, other: BitVector | int | Integer) -> bool: ...
    def __ne__(self, other: BitVector | int | Integer) -> bool: ...
    def __lt__(self, rhs: Signed | int | Integer) -> bool: ...
    def __gt__(self, rhs: Signed | int | Integer) -> bool: ...
    def __le__(self, rhs: Signed | int | Integer) -> bool: ...
    def __ge__(self, rhs: Signed | int | Integer) -> bool: ...
    def __lshift__(self, rhs: Unsigned | int | Integer) -> Signed: ...
    def __rshift__(self, rhs: Unsigned | int | Integer) -> Signed: ...
    def resize(self, target_width: int | None = None, *, zeros: int = 0) -> Signed: ...
